#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Utility and convenience functions for application-specific :mod:`setuptools`
subcommands.

Design
----------
This method intentionally duplicates existing utility functions provided by the
:mod:`betse.util` subpackage. While duplication breaks DRY ("Don't Repeat
Yourself") and hence is usually harmful, there are valid reasons to do so here.
Namely, :mod:`betse.util` functionality:

* Assumes BETSE to be available. While this is certainly the case when this
  file resides in the BETSE codebase, this is *not* necessarily the case when
  this file is copied into and hence resides in the codebases of other projects
  (e.g., BETSEE). In these projects, BETSE is merely yet another dependency
  that is typically unavailable at installation time.
* Raises BETSE-specific exceptions rooted at the BETSE-specific
  :class:`betse.exception.BetseException` superclass. :mod:`setuptools`
  subcommands, on the other hand, are expected to only raise
  :mod:`distutils`-specific exceptions rooted at the :mod:`distutils`-specific
  :class:`DistutilsError` superclass.
* Assumes logging to be configured. :mod:`setuptools`, however, assumes
  logging to *not* be configured -- and provides no assistance in doing so.
* Could theoretically import third-party dependencies unavailable at
  :mod:`setuptools subcommand time (e.g., due to the ``install`` or ``develop``
  subcommands *not* having been run yet). While no :mod:`betse.util` submodules
  should do so, the horrid possibility remains.

Since duplicating these functions here is no significant maintenance burden
*and* since attempting to reuse these functions here would introduce spurious
side effects, we adopt the former approach.
'''

# ....................{ IMPORTS                           }....................
import importlib, os, platform, shutil, subprocess, sys, time
from betsee import guimetadata
from distutils.errors import (
    DistutilsExecError, DistutilsFileError, DistutilsModuleError)
from os import path
from pkg_resources import Distribution, PathMetadata
from setuptools import Command

# ....................{ EXCEPTIONS ~ command              }....................
def die_unless_command_succeeds(*command_words) -> None:
    '''
    Raise an exception unless running the passed command succeeds.

    For portability, this command *must* be passed as a list of shell words
    whose first element is the pathname of such command and all subsequent
    elements the command-line arguments to be passed to such command (e.g.,
    ``['ls', '/']``).
    '''

    # If the first passed shell word is *NOT* pathable, raise an exception.
    die_unless_pathable(command_words[0])

    # Print the command to be run before doing so.
    print('Running "{}".'.format(' '.join(command_words)))

    # Keyword arguments to be passed to subprocess.check_call() below, which
    # accepts all arguments accepted by subprocess.Popen.__init__().
    popen_kwargs = {}

    # If the current platform is vanilla Windows, permit this command to
    # inherit all file handles (including stdin, stdout, and stderr) from the
    # current process. By default, subprocess.Popen documentation insists that:
    #
    #     "On Windows, if close_fds is true then no handles will be inherited
    #     by the child process."
    #
    # The child process will then open new file handles for stdin, stdout, and
    # stderr. If the current terminal is a Windows Console, the underlying
    # terminal devices and hence file handles will remain the same, in which
    # case this is *NOT* an issue. If the current terminal is Cygwin-based
    # (e.g.,, MinTTY), however, the underlying terminal devices and hence file
    # handles will differ, in which case this behaviour prevents interaction
    # between the current shell and the vanilla Windows command to be run
    # below.  In particular, all output from this command will be squelched.
    #
    # If at least one of stdin, stdout, or stderr are redirected to a blocking
    # pipe, setting "close_fds" to False can induce deadlocks under certain
    # edge-case scenarios. Since all such file handles default to None and
    # hence are *NOT* redirected in this case, "close_fds" may be safely set to
    # False.
    #
    # On all other platforms, if "close_fds" is True, no file handles *EXCEPT*
    # stdin, stdout, and stderr will be inherited by the child process. Hence,
    # this function fundamentally differs in subtle (and only slightly
    # documented ways) between vanilla Windows and all other platforms. These
    # discrepancies appear to be harmful but probably unavoidable, given the
    # philosophical gulf between vanilla Windows and all other platforms.
    if is_os_windows_vanilla():
        popen_kwargs['close_fds'] = False

    # Run this command.
    subprocess.check_call(command_words, **popen_kwargs)


def die_unless_pathable(command_basename: str, exception_message: str = None):
    '''
    Raise an exception with the passed message if the passed **pathable**
    (i.e., external command in the current ``${PATH}``) does *not* exist.
    '''

    # If this pathable is not found, raise an exception.
    if not is_pathable(command_basename):
        # If no message was passed, default this message.
        if not exception_message:
            exception_message = (
                'Command "{}" not found in the current ${{PATH}} or '
                'found but not an executable file.'.format(
                   command_basename))
        assert isinstance(exception_message, str), (
            '"{}" not a string.'.format(exception_message))

        # Raise this exception.
        raise DistutilsExecError(exception_message)

# ....................{ EXCEPTIONS ~ path                 }....................
def die_unless_basename(pathname: str, exception_message: str = None) -> None:
    '''
    Raise an exception unless the passed path is a **basename** (i.e., contains
    no platform-specific directory separator characters).
    '''

    # If this path is not a basename, fail.
    if not is_basename(pathname):
        # If no such message was passed, default this message.
        if not exception_message:
            exception_message = (
                'Pathname "{}" not a basename '
                '(i.e., either empty or '
                'contains a directory separator).'.format(pathname))
        assert isinstance(exception_message, str), (
            '"{}" not a string.'.format(exception_message))

        # Raise this exception.
        raise DistutilsFileError(exception_message)


def die_unless_dir_or_not_found(
    pathname: str, exception_message: str = None) -> None:
    '''
    Raise an exception unless the passed path is either an existing directory
    *or* does not exist (i.e., if this path is an existing non-directory).
    '''
    # If such path is an existing non-directory, fail.
    if is_path(pathname) and not is_dir(pathname):
        # If no such message was passed, default such message.
        if not exception_message:
            if is_file(pathname):
                exception_message =\
                    'Directory "{}" already an existing file.'.format(pathname)
            elif is_symlink(pathname):
                exception_message =\
                    'Directory "{}" already an existing symbolic link.'.format(
                        pathname)
            else:
                exception_message = 'Path "{}" not a directory.'.format(
                    pathname)
        assert isinstance(exception_message, str),\
            '"{}" not a string.'.format(exception_message)

        # Raise this exception.
        raise DistutilsFileError(exception_message)


def die_unless_file_or_not_found(
    pathname: str, exception_message: str = None) -> None:
    '''
    Raise an exception unless the passed path is either an existing non-special
    file _or_ does not exist (e.g., if such path is an existing directory).
    '''

    # If this path exists and is *NOT* an existing non-special file, fail.
    if is_path(pathname) and not is_file(pathname):
        # If no such message was passed, default this message.
        if not exception_message:
            if is_dir(pathname):
                exception_message = (
                    'File "{}" already an existing directory.'.format(
                        pathname))
            elif is_symlink(pathname):
                exception_message = (
                    'File "{}" already an existing symbolic link.'.format(
                        pathname))
            else:
                exception_message = 'Path "{}" not a file.'.format(pathname)
        assert isinstance(exception_message, str), (
            '"{}" not a string.'.format(exception_message))

        # Raise this exception.
        raise DistutilsFileError(exception_message)


def die_unless_path(pathname: str, exception_message: str = None) -> None:
    '''
    Raise an exception unless the passed path exists.
    '''

    # If this path is not found, fail.
    if not is_path(pathname):
        # If no such message was passed, default this message.
        if not exception_message:
            exception_message = 'Path "{}" not found.'.format(pathname)
        assert isinstance(exception_message, str), (
            '"{}" not a string.'.format(exception_message))

        # Raise this exception.
        raise DistutilsFileError(exception_message)


def die_unless_dir(dirname: str, exception_message: str = None) -> None:
    '''
    Raise an exception unless the passed directory exists.
    '''

    # If this directory is not found, fail.
    if not is_dir(dirname):
        # If no such message was passed, default such message.
        if not exception_message:
            exception_message = 'Directory "{}" not found.'.format(dirname)
        assert isinstance(exception_message, str), (
            '"{}" not a string.'.format(exception_message))

        # Raise this exception. Since there exists no
        # "DistutilsDirError" class, the next best thing is raised.
        raise DistutilsFileError(exception_message)


def die_unless_file(filename: str, exception_message: str = None) -> None:
    '''
    Raise an exception unless the passed non-special file exists.
    '''

    # If such file is not found, fail.
    if not is_file(filename):
        # If no such message was passed, default such message.
        if not exception_message:
            exception_message = 'File "{}" not found.'.format(filename)
        assert isinstance(exception_message, str), (
            '"{}" not a string.'.format(exception_message))

        # Raise this exception.
        raise DistutilsFileError(exception_message)


def die_unless_symlink(filename: str) -> None:
    '''
    Raise an exception unless the passed symbolic link exists.
    '''
    assert isinstance(filename, str), (
        '"{}" not a string.'.format(filename))

    if not is_symlink(filename):
        raise DistutilsFileError(
            'Symbolic link "{}" not found.'.format(filename))

# ....................{ EXCEPTIONS ~ python               }....................
def die_unless_module(module_name: str, exception_message: str = None):
    '''
    Raise an exception with the passed message if the module with the passed
    fully-qualified name (e.g., ``astarte.ashtoreth.ishtar``) is unimportable.
    '''

    # If this module is unimportable, raise an exception.
    if not is_module(module_name):
        # If no message was passed, default this message.
        if not exception_message:
            exception_message = (
                'Module "{}" not installed or not importable under '
                'the current Python interpreter.'.format(module_name))
        assert isinstance(exception_message, str), (
            '"{}" not a string.'.format(exception_message))

        # Raise this exception.
        raise DistutilsModuleError(exception_message)

# ....................{ TESTERS ~ os                      }....................
def is_os_linux() -> bool:
    '''
    `True` only if the current operating system is Linux.
    '''
    return platform.system() == 'Linux'


def is_os_posix() -> bool:
    '''
    ``True`` only if the current operating system complies with POSIX standards
    (e.g., as required for POSIX-compliant symbolic link support).

    Typically, this implies this system to *not* be vanilla Microsoft Windows
    (i.e., to be either a Cygwin-enabled Windows terminal *or* a genuine
    POSIX-compliant system).
    '''
    return os.name == 'posix'
    # return False


def is_os_os_x() -> bool:
    '''
    ``True`` only if the current operating system is Apple OS X.
    '''
    return platform.system() == 'Darwin'

# ....................{ TESTERS ~ os : windows            }....................
def is_os_windows() -> bool:
    '''
    ``True`` only if the current operating system is Microsoft Windows.

    This function reports ``True`` for both vanilla and Cygwin Microsoft
    Windows.
    '''
    return is_os_windows_vanilla() or is_os_windows_cygwin()


def is_os_windows_cygwin() -> bool:
    '''
    ``True`` only if the current operating system is **Cygwin Microsoft
    Windows** (i.e., running the Cygwin POSIX compatibility layer).
    '''
    return sys.platform == 'cygwin'


def is_os_windows_vanilla() -> bool:
    '''
    ``True`` only if the current operating system is **vanilla Microsoft
    Windows** (i.e., *not* running the Cygwin POSIX compatibility layer).
    '''
    return sys.platform == 'win32'

# ....................{ TESTERS ~ path                    }....................
def is_basename(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is a **basename** (i.e., is a non-empty
    string containing no platform-specific directory separator characters).
    '''
    assert isinstance(pathname, str), '"{}" not a string.'.format(pathname)

    return len(pathname) and pathname != path.basename(pathname)


def is_path(pathname: str) -> bool:
    '''
    ``True`` only if the passed path exists.
    '''
    assert isinstance(pathname, str), '"{}" not a string.'.format(pathname)

    return path.exists(pathname)


def is_dir(pathname: str) -> bool:
    '''
    ``True`` only if the passed directory exists.
    '''
    assert isinstance(pathname, str), '"{}" not a string.'.format(pathname)

    return path.isdir(pathname)


def is_file(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is an existing non-directory file exists
    *after* following symbolic links.

    Versus ``path.isfile()``
    ----------
    This function intrinsically differs from the standard :func:`path.isfile`
    function. While the latter returns ``True`` only for non-special files and
    hence ``False`` for all non-directory special files (e.g., device nodes,
    sockets), this function returns ``True`` for *all* non-directory files
    regardless of whether such files are special or not.

    **Why?** Because this function complies with POSIX semantics, whereas
    `path.isfile()` does *not*. The specialness of non-directory files is
    usually irrelevant; in general, it only matters whether such files are
    directories or not. For example, the external command ``rm`` removes only
    non-directory files (regardless of specialness) while the external command
    ``rmdir`` removes only empty directories.
    '''
    return is_path(pathname) and not is_dir(pathname)


def is_symlink(filename: str) -> bool:
    '''
    ``True`` only if the passed symbolic link exists.

    ``False`` is returned if the passed symbolic link exists but the current
    user has insufficient privelages to follow such link. This may constitute a
    bug in the underlying :func:`path.islink` function.
    '''
    assert isinstance(filename, str), '"{}" not a string.'.format(filename)
    return path.islink(filename)


def is_pathable(command_basename: str) -> bool:
    '''
    ``True`` only if the external command with the passed basename exists
    (i.e., is an executable file in the current ``${PATH}``).
    '''

    # Sanitize this command basename.
    command_basename = sanitize_command_basename(command_basename)

    # Return whether this command is found or not.
    return shutil.which(command_basename) is not None

# ....................{ TESTERS ~ module                  }....................
def is_module(module_name: str) -> bool:
    '''
    ``True`` only if the module with the passed fully-qualified name is
    importable under the active Python interpreter.

    If this module is a **submodule** (i.e., contains a ``.`` character), all
    parent modules of this module will be imported as a side effect of this
    function call. Likewise, if this module is *not* importable via standard
    mechanisms (e.g., the OS X-specific :mod:`PyObjCTools` package), the module
    itself may also be imported as a side effect.
    '''

    # See betse.util.python.modules.is_module() for implementation details.
    assert isinstance(module_name, str), (
        '"{}" not a string.'.format(module_name))
    assert len(module_name), 'Module name empty.'
    try:
        return importlib.util.find_spec(module_name) is not None
    except ValueError:
        try:
            importlib.import_module(module_name)
            return True
        except ImportError:
            return False

# ....................{ GETTERS ~ io                      }....................
def get_command_output(*args) -> str:
    '''
    Standard output and error captured by running the external shell command
    signified by the passed list if this command succeeds or raise an exception
    detailing this command's failure otherwise.

    Parameters
    ----------
    *args : list
        List of shell words comprising this command. The first item of this
        list should be the pathname for this command; all remaining items
        should be the arguments to pass this command.

    Returns
    ----------
    str
        All standard output and error captured by running this command,
        interleaved together in output order, stripped of all trailing
        newlines (as under most POSIX shells), *and* decoded via the current
        locale's preferred encoding (e.g., UTF-8).
    '''

    command_output = subprocess.check_output(
        args,

        # Redirect standard error to output.
        stderr=subprocess.STDOUT,

        # Decode this output using the current locale's preferred encoding.
        universal_newlines=True,
    )

    # Get this output, stripped of all trailing newlines.
    return command_output.rstrip('\n')

# ....................{ GETTERS ~ io : file               }....................
def get_chars(filename: str, encoding: str = 'utf-8') -> str:
    '''
    String of all characters contained in the plaintext file with the passed
    filename encoded with the passed encoding.

    Parameters
    ----------
    filename : str
        Relative or absolute path of the plaintext text to be read.
    encoding : optional[str]
        Name of the encoding to be used. Defaults to UTF-8.

    Returns
    ----------
    str
        String of all characters decoded from this file's byte content.
    '''
    assert isinstance(filename, str), '"{}" not a string.'.format(filename)
    assert isinstance(encoding, str), '"{}" not a string.'.format(encoding)

    with open(filename, mode='rt', encoding=encoding) as text_file:
        return text_file.read()

# ....................{ GETTERS ~ metadata                }....................
#FIXME: Reuse in the "betse_setup.betseutil" submodule as well.
def get_description() -> str:
    '''
    Human-readable multiline description of this application in
    reStructuredText (reST) format.

    To minimize synchronization woes, this description is identical to the
    contents of the :doc:`/README.rst` file. When submitting this application
    package to PyPI, this description is re-used verbatim as this package's
    front matter.

    Caveats
    ----------
    This function is I/O intensive and hence should be called sparingly --
    ideally, only once by this application's top-level ``setup.py`` script.
    '''

    # Relative path of this application's front-facing documentation in
    # reStructuredText format, required by PyPI. This path resides outside this
    # application's package tree and hence is inlined here rather than provided
    # by the "betsee.guipathtree" submodule.
    DESCRIPTION_FILENAME = 'README.rst'

    # Description read from this description file.
    try:
        description = get_chars(DESCRIPTION_FILENAME)
        # print('description: {}'.format(_DESCRIPTION))
    # If this file is *NOT* readable, print a non-fatal warning and reduce this
    # description to the empty string. While unfortunate, this description is
    # *NOT* required for most operations and hence mostly ignorable.
    except Exception as exception:
        description = ''
        output_warning(
            'Description file "{}" not found or not readable:\n{}'.format(
                DESCRIPTION_FILENAME, exception))

    # Retcurn this description.
    return description

# ....................{ GETTERS ~ path                    }....................
def get_project_dirname() -> str:
    '''
    Absolute path of the directory containing the currently run ``setup.py``
    script.
    '''

    # While such path is also typically  available as the first entry of the
    # "sys.path" list, you know what they say about assumptions.
    return get_path_dirname(get_path_dirname(__file__))


def get_path_canonicalized(pathname: str) -> str:
    '''
    **Canonical form** (i.e., unique absolute path) of the passed path.

    Specifically (in order):

    * Perform **tilde expansion,** replacing a `~` character prefixing this
      path by the absolute path of the current user's home directory.
    * Perform **path normalization,** thus:

      * Collapsing redundant separators (e.g., converting ``//`` to ``/``).
      * Converting relative to absolute path components (e.g., converting
        ``../`` to the name of the parent directory of such component).
    '''

    assert isinstance(pathname, str), '"{}" not a string.'.format(pathname)
    assert len(pathname), 'Pathname empty.'
    return path.abspath(path.expanduser(pathname))


def get_path_dirname(pathname: str) -> str:
    '''
    **Dirname** (i.e., parent directory) of the passed path if this path has a
    dirname *or* raise an exception otherwise.
    '''

    # Return this dirname. Since the path.dirname() function returns the empty
    # string for paths containing no directory separators and hence having no
    # dirnames, assert such return value to be non-empty.
    dirname = path.dirname(pathname)
    assert len(dirname), 'Pathname "{}" dirname empty.'.format(pathname)
    return dirname

# ....................{ GETTERS ~ path : filetype         }....................
def get_path_filetype(pathname: str) -> str:
    '''
    **Last filetype** (i.e., last ``.``-prefixed substring of the
    basename *not* including such ``.``) of the passed path if this path has a
    filetype *or* ``None`` otherwise.

    If this path contains multiple filetypes (e.g., ``odium.reigns.tar.gz``),
    this function returns only the last filetype.
    '''
    assert isinstance(pathname, str), '"{}" not a string.'.format(pathname)
    assert len(pathname), 'Pathname empty.'

    # Such filetype. (Yes, splitext() is exceedingly poorly named.)
    filetype = path.splitext(pathname)[1]

    # Get such filetype, stripping the prefixing "." from the string returned by
    # the prior call if such path has a filetype or returning None otherwise.
    return filetype[1:] if filetype else None


def get_path_sans_filetype(pathname: str) -> str:
    '''
    Get the passed path without last filetype (including prefixing ``.``) if
    this path has a filetype *or* as is otherwise.
    '''
    assert isinstance(pathname, str), '"{}" not a string.'.format(pathname)
    assert len(pathname), 'Pathname empty.'
    return path.splitext(pathname)[0]

# ....................{ SANITIZERS ~ metadata             }....................
#FIXME: Reuse in the "betse_setup.betseutil" submodule as well.
def sanitize_classifiers(classifiers: list) -> list:
    '''
    List of all PyPI-specific trove classifier strings synopsizing this
    application, manufactured by appending classifiers synopsizing this
    application's support for Python major versions (e.g.,
    ``Programming Language :: Python :: 3.6``, a classifier implying this
    application to successfully run under Python 3.6) to the passed list.
    '''
    assert isinstance(classifiers, list), '"{}" not a list.'.format(classifiers)

    # Major version of Python required by this application.
    PYTHON_VERSION_MAJOR = guimetadata.PYTHON_VERSION_MIN_PARTS[0]

    # List of classifiers to return, copied from the passed list for safety.
    classifiers_sane = classifiers[:]

    # For each minor version of Python 3.x supported by this application,
    # formally classify this version as such.
    for python_version_minor in range(
        guimetadata.PYTHON_VERSION_MIN_PARTS[1],
        guimetadata.PYTHON_VERSION_MINOR_MAX):
        classifiers.append(
            'Programming Language :: Python :: {}.{}'.format(
                PYTHON_VERSION_MAJOR, python_version_minor,))
    # print('classifiers: {}'.format(_CLASSIFIERS))

    # Return this sanitized list of classifiers.
    return classifiers_sane

# ....................{ SANITIZERS ~ path                 }....................
def sanitize_command_basename(command_basename: str) -> str:
    '''
    Convert the passed platform-agnostic command basename (e.g., ``pytest``)
    into a platform-specific command basename (e.g., ``pytest.exe``).

    If the passed basename contains a directory separator and hence is *not* a
    basename, an exception is raised. Else, under:

    * Windows, the passed basename is appended by ``.exe``. To avoid confusion
      with non-Windows executables in the current ``${PATH}`` when running under
      Wine emulation, only Windows executables are accepted when running under
      Windows.
    * All other platforms, the passed basename is returned as is.
    '''
    assert isinstance(command_basename, str), (
        '"{}" not a string.'.format(command_basename))
    assert len(command_basename), 'Command basename empty.'

    # If this pathname is *NOT* a basename, raise an exception.
    die_unless_basename(command_basename)

    # If this is Windows *AND* this basename has no filetype, suffix this
    # basename by ".exe".
    if is_os_windows() and get_path_filetype(command_basename) is None:
        # print('command basename "{}" filetype; {}'.format(
        #     command_basename, get_path_filetype(command_basename)))
        return command_basename + '.exe'

    # Else, return this basename as is.
    return command_basename

# ....................{ IMPORTERS                         }....................
def import_module(
    module_name: str, exception_message: str = None) -> type(sys):
    '''
    Dynamically import and return the module, package, or C extension with the
    passed fully-qualified name.

    If this module is unimportable, an exception with the passed message is
    raised.
    '''

    # If this module is unimportable, raise an exception.
    die_unless_module(module_name, exception_message)

    # Else, import and return this module.
    return importlib.import_module(module_name)

# ....................{ QUITTERS                          }....................
def exit_with_status(exit_status: int) -> None:
    '''
    Terminate the current Python process with the passed 0-based exit status.
    '''
    sys.exit(exit_status)

# ....................{ OUTPUTTERS                        }....................
def output_sans_newline(*strings) -> None:
    '''
    Print the passed strings to standard output *not* suffixed by a newline.

    By default, printed strings are suffixed by a newline.
    '''
    print(*strings, end = '')


def output_warning(*warnings) -> None:
    '''
    Print the passed warning messages to standard error.
    '''
    print('WARNING: ', *warnings, file=sys.stderr)

# ....................{ QUOTERS                           }....................
def shell_quote(text: str) -> str:
    '''
    Shell-quote the passed string.

    If the current operating system is:

    * *Not* Windows (e.g., Linux, OS X), the returned string is guaranteed to
      be suitable for passing as an arbitrary positional argument to external
      commands.
    * Windows, the returned string is suitable for passing *only* to external
      commands parsing arguments according in the same way as the Microsoft C
      runtime. Whereas *all* applications running under POSIX-compliant systems
      are required to parse arguments in the same manner (e.g., according to
      Bourne shell lexing), no such standard applies to applications running
      under Windows. For this reason, shell quoting is inherently unreliable
      under Windows.
    '''
    assert isinstance(text, str), '"{}" not a string.'.format(text)

    # If the current OS is vanilla Windows, do *NOT* perform POSIX-compatible
    # quoting. Vanilla Windows is POSIX-incompatible and hence does *NOT* parse
    # command-line arguments according to POSIX standards. In particular,
    # vanilla Windows does *NOT* treat single-quoted arguments as single
    # arguments but rather as multiple shell words delimited by the raw literal
    # `'`. This is circumventable by calling an officially undocumented
    # Windows-specific Python function. (Awesome.)
    if is_os_windows_vanilla():
        return subprocess.list2cmdline([text])
    # Else, perform POSIX-compatible quoting.
    else:
        import shlex
        return shlex.quote(text)

# ....................{ MAKERS                            }....................
def make_dir_unless_found(dirname: str) -> None:
    '''
    Create the passed directory if such directory does *not* already exist.

    All nonexistent parents of such directory will also be recursively created,
    mimicking the action of the conventional shell command ``mkdir -p``.
    '''
    assert isinstance(dirname, str), '"{}" not a string.'.format(dirname)
    assert len(dirname), 'Dirname empty.'

    # If such directory does *NOT* already exist, create such directory. To
    # support logging, such condition is explicitly tested for. To avoid race
    # conditions (e.g., in the event such directory is created between testing
    # and creating such directory), we preserve the makedirs() keyword argument
    # "exist_ok = True".
    if not is_dir(dirname):
        # Log such creation.
        print('Creating directory "{}".'.format(dirname))

        # Create such directory if still needed.
        os.makedirs(dirname, exist_ok = True)


def make_symlink(pathname_source: str, filename_target: str) -> None:
    '''
    Symbolically link the passed source path to the passed target symlink.

    If this target is an existing symlink, this symlink will be implicitly
    removed before being recreated.

    If this source does *not* exist, an exception will be raised. Hence, this
    function does *not* support creation of **dangling symbolic links** (i.e.,
    links to non-existent paths).
    '''

    # If such source path does *NOT* exist, raise an exception.
    die_unless_path(pathname_source)

    # If such link currently exists, remove such link.
    if is_symlink(filename_target):
        remove_symlink(filename_target)

    # (Re)create such link.
    print('Symbolically linking "{}" to "{}".'.format(
        pathname_source, filename_target))
    os.symlink(pathname_source, filename_target)

# ....................{ MOVERS                            }....................
def move_file(filename_source: str, filename_target: str) -> None:
    '''
    Move the passed source to the passed target file.
    '''

    # If such file does *NOT* exist, raise an exception.
    die_unless_file(filename_source)

    # Move such file.
    print('Moving file "{}" to "{}".'.format(filename_source, filename_target))
    shutil.move(filename_source, filename_target)

# ....................{ REMOVERS                          }....................
def remove_path(pathname: str) -> None:
    '''
    Recursively remove the passed directory in a safe manner (e.g., *not*
    following symbolic links outside such directory).

    This is an inherently dangerous operation and hence delayed for several
    seconds, allowing sufficiently aware users to jam the panic button.
    '''

    # If such path does *NOT* exist, fail.
    die_unless_path(pathname)

    # If such path is a directory, remove such directory.
    if is_dir(pathname):
        remove_dir(pathname)
    # Else, remove such file.
    else:
        remove_file(pathname)


def remove_dir(dirname: str) -> None:
    '''
    Recursively remove the passed directory in a safe manner (e.g., *not*
    following symbolic links outside such directory).

    This is an inherently dangerous operation and hence delayed for several
    seconds, allowing sufficiently aware users to jam the panic button.
    '''

    # If such directory does *NOT* exist, fail.
    die_unless_dir(dirname)

    # For safety, wait several seconds to do so. (Read: panic button.)
    sleep_seconds = 4
    print('Removing directory "{}" in {} seconds...'.format(
        dirname, sleep_seconds))
    time.sleep(sleep_seconds)

    # Remove such directory.
    shutil.rmtree(dirname)
    print('Removed directory "{}".'.format(dirname))


def remove_file(filename: str) -> None:
    '''
    Remove the passed non-special file.
    '''

    # If such file does *NOT* exist, fail.
    die_unless_file(filename)

    # Remove such file.
    print('Removing file "{}".'.format(filename))
    os.unlink(filename)


def remove_symlink(filename: str) -> None:
    '''
    Remove the passed symbolic link.
    '''

    # If this link does *NOT* exist, fail.
    die_unless_symlink(filename)

    # Remove this link. Since symbolic links are special files, remove_file()
    # fails when passed such link and hence must be reimplemented here.
    print('Removing symbolic link "{}".'.format(filename))
    os.unlink(filename)

# ....................{ SETUPTOOLS                        }....................
def add_setup_command_classes(
    metadata: dict, setup_options: dict, *command_classes) -> None:
    '''
    Add one application-specific :mod:`setuptools` command for each passed
    class to the passed dictionary of :mod:`setuptools` options.

    For simplicity, the name of each such command will be the name of the
    corresponding class. Hence, the names of such classes are recommended to be
    short lowercase strings (e.g., ``freeze``, ``symlink``).
    '''
    assert isinstance(metadata, dict), (
        '"{}" not a dictionary.'.format(metadata))
    assert isinstance(setup_options, dict), (
        '"{}" not a dictionary.'.format(setup_options))

    # Add each such command class as a new command of the same name.
    for command_class in command_classes:
        assert isinstance(command_class, type), (
            '"{}" not a class.'.format(command_class))

        # Add this command.
        setup_options['cmdclass'][command_class.__name__] = command_class

        # Expose the passed dictionaries to this class by monkey-patching
        # application-specific private class variables into these classes.
        # While passing these dictionaries to instances of this class (e.g., on
        # instantiation) would be ideal, distutils and hence setuptools
        # requires commands to be registered as classes rather than instances.
        command_class._metadata = metadata
        command_class._setup_options = setup_options

# ....................{ SETUPTOOLS ~ wrapper : generator  }....................
def command_entry_points(command: Command) -> 'GeneratorType':
    '''
    Generator yielding a 3-tuple detailing each wrapper script installed for
    the **Python distribution** (i.e., top-level package) identified by the
    passed `setuptools` command.

    See Also
    ----------
    `package_distribution_entry_points()`
        Further details on tuple contents.
    '''
    assert isinstance(command, Command), (
        '"{}" not a setuptools command.'.format(command))

    # Make a "pkg_resources"-specific distribution from the passed command.
    # Yes, this code was ripped wholesale from the run() method defined by the
    # "setuptools.command.install_scripts" submodule. Yes, we don't know how it
    # works. ("Frankly, Mam, we don't give a damn.")
    #
    # It should be noted that all commands have an attribute "distribution".
    # Naturally, this is a setuptools-specific distribution that has literally
    # *NOTHING* to do with pkg_resources-style distributions.
    #
    # Die, setuptools. Die!
    ei_cmd = command.get_finalized_command('egg_info')
    distribution = Distribution(
        ei_cmd.egg_base,
        PathMetadata(ei_cmd.egg_base, ei_cmd.egg_info),
        ei_cmd.egg_name,
        ei_cmd.egg_version,
    )

    # Defer to the generator provided by this function.
    yield from package_distribution_entry_points(distribution)


def package_distribution_entry_points(
    distribution: '(Distribution, VersionlessRequirement)') -> None:
    '''
    Generator yielding a 3-tuple describing each wrapper script installed for
    the passed `pkg_resources`-specific distribution identifying a unique
    top-level Python package.

    Yields
    ----------
    script_basename, ui_type, entry_point
        3-tuple such that:

        * ``script_basename` is this script's basename (e.g., ``betse``). To
          simplify integration with the downstream setuptools API (e.g., the
          :meth:`setuptools.command.easy_install.ScriptWriter.get_script_args`
          method, this basename is typically *not* suffixed by a
          platform-specific filetype (e.g., ``.exe`` under vanilla or Cygwin
          Microsoft Windows).
        * ``ui_type`` is this script's interface type string, guaranteed to be
          either:

          * ``console`` if this script is console-specific.
          * ``gui`` otherwise.

        * ``entry_point`` is this script's :class:`EntryPoint` object, whose
          attributes specify the module to be imported and function to be run
          by this script.

    Parameters
    ----------
    distribution : Distribution, VersionlessRequirement
        Distribution object identifying the top-level Python package to yield
        entry points for. Specifically, either:

        * A :class:`Distribution` object supplied by the ``install`` or
          ``symlink`` subcommands.
        * A :class:`VersionlessRequirement` object supplied by the ``develop``
          subcommand. As the classname suggests, this object wraps the
          corresponding :class:`Distribution` object by stripping versioning
          from this distribution's name (e.g., reducing ``foo==1.0`` to merely
          ``foo``).
    '''
    # Do *NOT* bother attempting to assert the passed distribution to be an
    # instance of either the "Distribution" or "VersionlessRequirement"
    # classes.  While the former is guaranteed to exist under all setuptools
    # version, the latter is *NOT*. Instead, only assert this distribution to
    # be non-None.
    assert distribution is not None, 'Setuptools distribution expected.'

    # For each type of script wrapper...
    for script_type in 'console', 'gui':
        script_type_group = script_type + '_scripts'

        # For each script of this type...
        for script_basename, entry_point in (
            distribution.get_entry_map(script_type_group).items()):
            # Yield this 3-tuple. To simplify integration with the downstream
            # setuptools API, do *NOT* sanitize_snakecase this script's
            # basename by calling sanitize_command_basename(). Since that API
            # already implicitly suffixes this basename by ".exe", doing so
            # here would erroneously result in this basename being suffixed by
            # ".exe.exe".
            yield script_basename, script_type, entry_point



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level :mod:`PySide2`-based application-wide settings facilities.

See Also
----------
:class:`PySide2.QtCore.QSettings`
    This class' official documentation is a comprehensive, comprehensible
    commentary on cross-platform, thread- and process-safe (de)serialization of
    application-wide settings. This documentation doubles as a human-readable
    FAQ and hence comes recommended, particularly for the Qt neophyte.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QSettings
from betse.util.os import oses

# ....................{ INITIALIZERS                       }....................
def init() -> None:
    '''
    Initialize the :class:`QSettings` class *before* the :func:`make` function
    is called elsewhere to instantiate that class.

    Specifically, this function establishes the default settings format
    subsequently accessed by the default :class:`QSettings` constructor as
    follows:

    * If the current platform is non-Cygwin Windows, request that settings be
      formatted in INI format to a physical file. By default, Windows settings
      are formatted as registery keys. Since the registery is fragile, insecure,
      and broken by design, this default remains unfortunate but fixable.
    * Else, request that settings be formatted in the format preferred by the
      current platform guaranteed to be a physical file.
    '''

    # Set the default settings format in a platform-specific manner.
    QSettings.setDefaultFormat(
        QSettings.IniFormat if oses.is_windows_vanilla() else
        QSettings.NativeFormat)

# ....................{ MAKERS                             }....................
def make() -> QSettings:
    '''
    Create and return a new :class:`QSettings` instance, encapsulating all
    application-wide settings in a cross-platform, thread- and process-safe
    manner guaranteed to be safely (de)serializable to and from a predefined
    application- and user-specific on-disk file.

    Design
    ----------
    Note that repeatedly calling this function throughout the lifetime of this
    application is guaranteed to be efficient by design. Thus, callers are
    encouraged to do so rather than persist a permanent reference to the first
    :class:`QSettings` instance returned by this function.
    '''

    # Default settings leveraging the default QSettings() constructor,
    # including:
    #
    # * The "QCoreApplication.applicationName", "organizationName", and
    #   "organizationName" attributes previously set by the "betsee.util.psdapp"
    #   submodule.
    # * The "QSettings.defaultFormat" attribute previously set by the above
    #   init() function.
    # * The default "QSettings.UserScope" scope.
    settings = QSettings()

    # Prevent settings from being read and/or written in a Cascading Style
    # Sheets (CSS)-style manner over a hierarchically nested multiplicity of
    # discontiguous files littered across the filesystem.
    #
    # Equivalently, read and write settings only to the single file specific to
    # the current platform, application, and user.
    settings.setFallbacksEnabled(False)

    # Encode settings persisted to INI files (e.g., on *nix platforms) with the
    # standard UTF-8. By default, settings are persisted to such files with the
    # non-standard INI escape sequences "codec." That's just asking for trouble.
    settings.setIniCodec('UTF-8')

    # Return these settings.
    return settings



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **simulator state** (i.e., abstract node in the finite state automata
corresponding to this simulator) functionality.
'''

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import QCoreApplication
from betse.science.pipe.export.pipeexpenum import SimExportType
from betse.science.phase.phaseenum import SimPhaseKind
# from betse.util.type.types import type_check
from betsee.gui.simtab.run.guisimrunenum import SimmerState, SimmerModelState

# ....................{ GLOBALS ~ dict                    }....................
#FIXME: This absolutely does *NOT* belong here. Shift this into a higher-level
#and more general-purpose submodule. This is sufficiently generic.
SIM_PHASE_KIND_TO_NAME = {
    SimPhaseKind.SEED: QCoreApplication.translate(
        'guisimrunstate', 'seed'),
    SimPhaseKind.INIT: QCoreApplication.translate(
        'guisimrunstate', 'initialization'),
    SimPhaseKind.SIM: QCoreApplication.translate(
        'guisimrunstate', 'simulation'),
}
'''
Dictionary mapping from each type of simulation phase to the human-readable
translated name of that phase.
'''

# ....................{ GLOBALS ~ set                     }....................
SIMMER_STATES = set(simmer_state for simmer_state in SimmerState)
'''
Set of all possible simulator states.
'''


SIMMER_STATES_IDLE = {
    SimmerState.UNQUEUED,
    SimmerState.QUEUED,
}
'''
Set of all **idle simulator states** (i.e., states implying *no* queued
subcommands to be currently working, stopping, or recently finished).
'''

# ....................{ GLOBALS ~ set : running           }....................
SIMMER_STATES_RUNNING = {
    SimmerState.MODELLING,
    SimmerState.EXPORTING,
}
'''
Set of all **running simulator states** (i.e., states implying one or more
queued subcommands to be currently running and hence either modelling or
exportingb but neither paused, stopping, nor finished).
'''


SIMMER_STATES_WORKING = SIMMER_STATES_RUNNING | {SimmerState.PAUSED,}
'''
Set of all **working simulator states** (i.e., states implying one or more
queued subcommands to be currently working and hence either modelling,
exporting, or paused but neither stopping nor finished).
'''

# ....................{ GLOBALS ~ set : halting           }....................
SIMMER_STATES_HALTING = {
    SimmerState.PAUSED,
    SimmerState.STOPPING,
    SimmerState.FINISHED,
}
'''
Set of all **halting simulator states** (i.e., states implying one or more
previously running queued subcommands to be currently halted, either
temporarily as in the case of a paused subcommand or permanently as in the
case of a stopped or finished subcommand).
'''


SIMMER_STATES_UNWORKABLE = {
    SimmerState.UNQUEUED,
    SimmerState.STOPPING,
}
'''
Set of all **unworkable simulator states** (i.e., states implying the proactor
to be incapable of performing any work).
'''

# ....................{ GLOBALS ~ set : from              }....................
SIMMER_STATES_FROM_FLUID = {
    SimmerState.UNQUEUED,
    SimmerState.QUEUED,
    SimmerState.FINISHED,
    SimmerState.STOPPING,
}
'''
Set of all **out-bound fluid simulator states** (i.e., states the simulator may
freely transition from into any other state).

Specifically, the states:

* Contained in this set are all low priority and hence may be flexibly replaced
  at simulator runtime with any other state.
* Excluded by this set are all high priority and hence *must* be preserved as
  is until completed -- possibly by manual user intervention (e.g., toggling
  the start and pause buttons).

Caveats
----------
Technically, the claims implied by this set are *not* actually all the case.
For example, the simulator may transition from the:

* :attr:`SimmerState.UNQUEUED` state to *only* the
  :attr:`SimmerState.QUEUED` state.
* :attr:`SimmerState.QUEUED` state to *only* the
  :attr:`SimmerState.UNQUEUED`,
  :attr:`SimmerState.MODELLING`, and
  :attr:`SimmerState.EXPORTING` states.

Pragmatically, preserving these distinctions would incur more cost than
pretending these states may be freely transitioned from. We prefer the latter.

See Also
----------
:data:`SIMMER_STATES_FROM_FIXED`
    Complement (i.e., negation) of this set.
'''


SIMMER_STATES_FROM_FIXED = SIMMER_STATES - SIMMER_STATES_FROM_FLUID
'''
Set of all **out-bound fixed simulator states** (i.e., states the simulator may
*not* freely transition from into any other state).

Specifically, the states:

* Contained in this set are all high priority and hence may be inflexibly
  replaced at simulator runtime with *only* other states in this same set.
* Excluded by this set are all low priority.

See Also
----------
:data:`SIMMER_STATES_FROM_FLUID`
    Complement (i.e., negation) of this set.
'''

# ....................{ GLOBALS ~ set : into              }....................
SIMMER_STATES_INTO_FLUID = {
    SimmerState.UNQUEUED,
    SimmerState.QUEUED,
}
'''
Set of all **in-bound fluid simulator states** (i.e., states the simulator may
freely transition into from any other state).

See Also
----------
:data:`SIMMER_STATES_FROM_FLUID`
    Analogue of this set with respect to in-bound transitions.
:data:`SIMMER_STATES_INTO_FIXED`
    Complement (i.e., negation) of this set.
'''


SIMMER_STATES_INTO_FIXED = SIMMER_STATES - SIMMER_STATES_INTO_FLUID
'''
Set of all **in-bound fixed simulator states** (i.e., states the simulator may
*not* freely transition into from any other state).

See Also
----------
:data:`SIMMER_STATES_FROM_FIXED`
    Analogue of this set with respect to in-bound transitions.
:data:`SIMMER_STATES_INTO_FLUID`
    Complement (i.e., negation) of this set.
'''

# ....................{ GLOBALS ~ dict : status           }....................
SIMMER_STATE_TO_PHASE_STATUS = {
    SimmerState.UNQUEUED: QCoreApplication.translate(
        'guisimrunstate', 'Unqueued'),
    SimmerState.QUEUED: QCoreApplication.translate(
        'guisimrunstate', 'Queued'),
    SimmerState.MODELLING: QCoreApplication.translate(
        'guisimrunstate', 'Modelling'),
    SimmerState.EXPORTING: QCoreApplication.translate(
        'guisimrunstate', 'Exporting'),
    SimmerState.PAUSED: QCoreApplication.translate(
        'guisimrunstate', 'Paused'),
    SimmerState.STOPPING: QCoreApplication.translate(
        'guisimrunstate', 'Finishing'),
    SimmerState.FINISHED: QCoreApplication.translate(
        'guisimrunstate', 'Finished'),
}
'''
Dictionary mapping from each type of simulator state to a human-readable,
translated string serving as a single-word synopsis of the action being
performed in that state.
'''


SIMMER_STATE_TO_PROACTOR_STATUS = {
    SimmerState.UNQUEUED: QCoreApplication.translate(
        'guisimrunstate', 'Waiting for phase(s) to be queued...'),
    SimmerState.QUEUED: QCoreApplication.translate(
        'guisimrunstate', 'Waiting for queued phase(s) to be started...'),
    SimmerState.MODELLING: QCoreApplication.translate(
        'guisimrunstate', 'Modelling <b>{phase_type}</b> phase...'),
    SimmerState.EXPORTING: QCoreApplication.translate(
        'guisimrunstate', 'Exporting <b>{phase_type}</b> phase...'),
    SimmerState.PAUSED: QCoreApplication.translate(
        'guisimrunstate', 'Paused {status_prior}.'),
    SimmerState.STOPPING: QCoreApplication.translate(
        'guisimrunstate', 'Finishing {status_prior}...'),
    SimmerState.FINISHED: QCoreApplication.translate(
        'guisimrunstate', 'Finished {status_prior}.'),
}
'''
Dictionary mapping from each type of simulator state to a human-readable,
translated, unformatted string templating a single-sentence synopsis of the
action being performed in that state.

Formats
----------
Format specifiers embedded in these strings include:

* ``{phase_type}``, a word signifying the type of currently running simulator
  phase if any (e.g., "seed," "initialization").
* ``{status_prior}``, text previously formatted from a string of this
  dictionary synopsizing the prior state of this simulator. Since this text is
  interpolated into the middle of arbitrary sentences, the first character of
  this text *must* be lowercase.

Most such strings contain no format specifiers and are thus displayable as is.
Some such strings contain one or more format specifiers (e.g., ``{cmd_name}}`)
and are thus displayable *only* after interpolating the corresponding values.
'''

# ....................{ GLOBALS ~ dict : substatus        }....................
SIMMER_STATE_TO_PROACTOR_SUBSTATUS = {
    SimmerState.UNQUEUED: QCoreApplication.translate(
        'guisimrunstate',
        'Queued '
        '{queued_modelling} phase(s) for modelling and '
        '{queued_exporting} phase(s) for exporting.'
    ),
    SimmerState.QUEUED: QCoreApplication.translate(
        'guisimrunstate',
        'Queued '
        '{queued_modelling} phase(s) for modelling and '
        '{queued_exporting} phase(s) for exporting.'
    ),
    SimmerState.MODELLING: {
        SimPhaseKind.SEED: None,
        SimPhaseKind.INIT: {
            SimmerModelState.PREPARING: QCoreApplication.translate(
                'guisimrunstate', 'Loading seeded cell cluster...'),
            SimmerModelState.MODELLING: QCoreApplication.translate(
                'guisimrunstate',
                'Initialized '
                '<b>{progress_current}</b> <i>of</i> '
                '<b>{progress_maximum}</b> '
                'simulation time steps.'
            ),
            SimmerModelState.FINISHING: QCoreApplication.translate(
                'guisimrunstate', 'Saving initialization results...'),
        },
        SimPhaseKind.SIM: {
            SimmerModelState.PREPARING: QCoreApplication.translate(
                'guisimrunstate', 'Loading initialization results...'),
            SimmerModelState.MODELLING: QCoreApplication.translate(
                'guisimrunstate',
                'Simulated '
                '<b>{progress_current}</b> <i>of</i> '
                '<b>{progress_maximum}</b> '
                'simulation time steps.'
            ),
            SimmerModelState.FINISHING: QCoreApplication.translate(
                'guisimrunstate', 'Saving simulation results...'),
        },
    },
    #FIXME: Consider querying the BETSE client for the metadata interpolated
    #into the following strings by defining the following new worker signals:
    #
    #* subprogress_ranged(), enabling us to at least subdivide a unit of
    #  progress (e.g., exportation of a single animation) into subunits of
    #  subprogress (e.g., each frame of that animation).
    #* subprogressed(), analogous to the existing progressed() signal.
    #* Some variant of pathname_saved(), pathname_wrote(), pathname_saving(),
    #  pathname_writing(), or pathname_written(), accepting a single string
    #  pathname. The saved variants read a tad better than the wrote variants,
    #  but none of these stand out as particularly inspiring. Tense is
    #  certainly an issue, as we'd rather callers not assume this pathname to
    #  have already been written to by this worker when this signal is emitted.
    SimmerState.EXPORTING: {
        SimExportType.CSV: QCoreApplication.translate(
            'guisimrunstate',
            'Exported comma-separated value (CSV) file '
            '<pre>"{pathname}"</pre>.'
        ),
        SimExportType.PLOT: QCoreApplication.translate(
            'guisimrunstate',
            'Exported image <pre>"{pathname}"</pre>.'
        ),
        SimExportType.ANIM: QCoreApplication.translate(
            'guisimrunstate',
            #FIXME: Replace this less descriptive synopsis with the more
            #verbose synopsis given below *AFTER* adding support for the
            #"subprogress_current" and "subprogress_total" fields. (See above.)
            'Exported animation <pre>"{pathname}"</pre>.'
            # 'Exported animation <pre>"{pathname}"</pre> frame '
            # '{subprogress_current} <i>of</i> {subprogress_total}.'
        ),
    },
    SimmerState.PAUSED:   '{substatus_prior}',
    SimmerState.STOPPING: '{substatus_prior}',
    SimmerState.FINISHED: '{substatus_prior}',
}
'''
Dictionary mapping from each type of simulator state to an object specific to
that type transitively yielding a human-readable, translated, unformatted
string templating the single-sentence details of the action being performed in
that state.

Structure
----------
Whereas each value of each key-value pair of the comparable
:data:`SIMMER_STATE_TO_PHASE_STATUS` and
:data:`SIMMER_STATE_TO_PROACTOR_STATUS` dictionaries is *always* a string,
each value of each key-value pair of this dictionary contextually depends upon
that value. Specifically, if this key is:

* :attr:`SimmerState.MODELLING`, this value is a nested dictionary mapping from
  each type of simulation phase to either:

  * If that phase is the seed phase, ``None``. Due to the variability of
    details pertaining to this phase, BETSE itself defines these details by the
    :meth:`betsee.gui.simtab.run.work.guisimrunworksig.SimCallbacksSignaller.progress_stated`
    method emitting a signal connected to the text box showing these details.
  * If that phase is the initialization or simulation phase, a nested
    dictionary mapping from each type of simulator modelling state to an
    unformatted string template detailing the current action being performed in
    that state.

* :attr:`SimmerState.EXPORTING`, this value is a nested dictionary mapping from
  each type of simulation export to an unformatted string template detailing
  the exporting of that export.
* Any other key, this value is an unformatted string template.

Formats
----------
Format specifiers embedded in these strings include:

* ``{progress_current}``, a non-negative integer signifying the current time
  step of this phase.
* ``{progress_total}``, a non-negative integer signifying the total number of
  time steps in this phase.

Most such strings contain no format specifiers and are thus displayable as is.
Some such strings contain one or more format specifiers (e.g., ``{cmd_name}}`)
and are thus displayable *only* after interpolating the corresponding values.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all editable scalar simulation configuration widget
subclasses instantiated in pages of the top-level stack.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import Signal, Slot  # QCoreApplication
from PySide2.QtWidgets import QUndoCommand
from betse.exceptions import BetseMethodUnimplementedException
from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.guiexception import BetseePySideWidgetException
from betsee.gui.simconf.stack.widget.abc.guisimconfwdgedit import (
    QBetseeSimConfEditWidgetMixin)
from betsee.util.widget.abc.guiundocmdabc import QBetseeWidgetUndoCommandABC

# ....................{ MIXINS                             }....................
class QBetseeSimConfEditScalarWidgetMixin(QBetseeSimConfEditWidgetMixin):
    '''
    Abstract base class of all **editable scalar simulation configuration
    widget** (i.e., widget interactively editing scalar simulation configuration
    values stored in external YAML files) subclasses.

    In this context, the term "scalar" encompasses all widget subclasses whose
    contents reduce to a single displayed value (e.g., integer, floating point
    number, string).

    Attributes
    ----------
    _widget_value_last : object
        Previously displayed value of this widget cached on the completion of
        the most recent user edit (i.e., :meth:`editingFinished` signal),
        possibly but *not* necessarily reflecting this widget's current state.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this editable scalar widget mixin.
        '''

        # Initialize our superclass with all passed arguments.
        super().__init__(*args, **kwargs)

        # Nullify all remaining instance variables for safety.
        self._widget_value_last = None


    def init(self, *args, **kwargs) -> bool:

        # Initialize our superclass with all passed parameters.
        super().init(*args, **kwargs)

        # Connect all relevant signals to slots *AFTER* initializing our
        # superclass. See the superclass method for details.
        #
        # On each finalized change to the value displayed by this widget, set
        # the current value of the simulation configuration alias associated
        # with this widget to this widget's displayed value.
        self._finalize_widget_change_signal.connect(
            self._set_alias_to_widget_value_if_sim_conf_open)

    # ..................{ SUBCLASS ~ mandatory : property    }..................
    # Subclasses are required to implement the following properties.

    @property
    def undo_synopsis(self) -> str:
        '''
        Human-readable string synopsizing the operation performed by this scalar
        widget, preferably as a single translated sentence fragment.
        '''

        raise BetseMethodUnimplementedException()


    @property
    def widget_value(self) -> object:
        '''
        High-level :mod:`PySide2`-specific scalar value currently displayed by
        this scalar widget.

        Each subclass typically implements this Python property in terms of an
        unprefixed getter method of this widget (e.g., :meth:`QLineEdit.text`).

        Caveats
        ----------
        If this value is neither of the exact type(s) required by the simulation
        configuration alias associated with this widget *nor* of a similar type
        safely convertible into such a type, the subclass *must* redefine both
        the :meth:`_get_alias_from_widget_value` and
        :meth:`_get_widget_from_alias_value` methods to convert this value to
        and from such a type.

        This high-level value is purely :mod:`PySide2`-specific and hence
        distinct from the associated low-level scalar value defined by the
        simulation configuration. In particular, these two values are typically
        but *not* necessarily of the same type.

        For example, for the :class:`QBetseeSimConfEnumComboBox` subclass:

        * The high-level :meth:`QBetseeSimConfEnumComboBox.widget_value`
          property returns an integer (i.e., the 0-based index of the currently
          selected item in that combo box).
        * The low-level :meth:`QBetseeSimConfEnumComboBox._sim_conf_alias.get`
          getter returns the enumeration member corresponding to this item.
        '''

        raise BetseMethodUnimplementedException()


    @widget_value.setter
    def widget_value(self, widget_value: object) -> None:
        '''
        Set the high-level :mod:`PySide2`-specific scalar value currently
        displayed by this scalar widget to the passed value.

        Each subclass typically implements this Python property in terms of a
        ``set``-prefixed setter method of this widget (e.g.,
        :meth:`QLineEdit.setText`).

        Caveats
        ----------
        To avoid infinite recursion, the superclass rather than subclass
        implementation of this setter method should typically be called
        (e.g., ``super().setText()`` rather than ``self.setText()``). For the
        :class:`QBetseeSimConfLineEdit` subclass, for example, erroneously
        calling this subclass implementation would ensure that:

        #. On each call to the :meth:`setValue` method...
        #. Which pushes an undo command onto the undo stack...
        #. Whose :meth:`QUndoCommand.redo` method is called by that stack...
        #. Which calls the :meth:`setValue` method...
        #. Which induces infinite recursion.
        '''

        raise BetseMethodUnimplementedException()


    @property
    def _finalize_widget_change_signal(self) -> Signal:
        '''
        Signal signalled on each finalized interactive user (but *not*
        programmatic) edit of the contents of this widget.

        The :meth:`init` method implicitly connects this signal to the
        :meth:`_set_alias_to_widget_value_if_sim_conf_open` slot.
        '''

        raise BetseMethodUnimplementedException()

    # ..................{ SUBCLASS ~ mandatory : method      }..................
    # Subclasses are required to implement the following methods.

    def _reset_widget_value(self) -> None:
        '''
        Reset the scalar value currently displayed by this scalar widget, thus
        reverting this widget to its default state divorced from an underlying
        model.

        For example, if this widget displays:

        * A string value, this method should set this value to the empty string.
        * A float value, this method should set this value to 0.0.
        * An integer value, this method should set this value to 0.
        '''

        raise BetseMethodUnimplementedException()

    # ..................{ SLOTS                              }..................
    @Slot(str)
    def _set_filename(self, filename: str) -> None:

        # Call the superclass method first.
        super()._set_filename(filename)

        # Set the value displayed by this widget to the current value of this
        # simulation configuration alias.
        self._set_widget_to_alias_value(filename)

    # ..................{ CONVERTERS ~ alias -> widget       }..................
    # Called on opening and closing simulation configurations.
    @type_check
    def _set_widget_to_alias_value(self, filename: str) -> None:
        '''
        Set this widget's displayed value to the current value of the
        simulation configuration alias associated with this widget if a
        simulation configuration is currently open *or* clear this displayed
        value otherwise.

        This method is currently only called by the :meth:`_set_filename` slot
        on opening and closing a simulation configuration, thus initializing
        this widget's value to that of this configuration.

        Parameters
        ----------
        filename : str
            Absolute path of the currently open YAML-formatted simulation
            configuration file if any *or* the empty string otherwise (i.e., if
            no such file is open).
        '''

        # If a simulation configuration is currently open...
        if filename and self._is_sim_open:
            # Set this widget's displayed value from this alias' current value.
            self.widget_value = self._get_widget_from_alias_value()

            # If no such value is returned, (e.g., due to a subclass improperly
            # overriding the _get_widget_from_alias_value() function), raise an
            # exception. *NO* simulation configuration alias currently returns
            # "None" as a valid value.
            if self.widget_value is None:
                raise BetseePySideWidgetException(
                    'Editable scalar widget "{}" value "None" invalid.'.format(
                        self.obj_name))

            # Log this setting.
            logs.log_debug(
                'Setting widget "%s" display value to %r...',
                self.obj_name, self.widget_value)

            # Cache this widget's value in preparation for the next change.
            self._widget_value_last = self.widget_value
        # Else, no simulation configuration is currently open.
        else:
            # Clear this widget's displayed and cached values.
            self._reset_widget_value()
            self._widget_value_last = None


    def _get_widget_from_alias_value(self) -> object:
        '''
        Value of the simulation configuration alias associated with this widget,
        coerced into a type displayable by this widget.

        This method is typically called *only* once per open simulation
        configuration file on the first loading of that file.

        See Also
        ----------
        :meth:`_get_alias_from_widget_value`
            Further details.
        '''

        return self._sim_conf_alias.get()

    # ..................{ CONVERTERS ~ widget -> alias       }..................
    # Called on each user edit of this widget's value.
    @Slot()
    def _set_alias_to_widget_value_if_sim_conf_open(self) -> None:
        '''
        Slot signalled on each finalized interactive user edit (and possibly but
        *not* necessarily each non-interactive programmatic change) to the
        value displayed by this widget, setting the current value of the
        simulation configuration alias associated with this widget to this
        widget's displayed value if a simulation configuration is currently open
        *or* reduce to a noop otherwise.

        Design
        ----------
        If Qt signals the :meth:`_finalize_widget_change_signal` connected to
        this slot *only* on each finalized interactive user edit (rather than
        both each such edit *and* each non-interactive programmatic change),
        the subclass *must* explicitly call this method on each non-interactive
        programmatic change -- typically in the subclass implementation of this
        widget's main setter method (e.g., :meth:`QLineEdit.setText`),
        guaranteed to be called on each such change.

        While *not* necessarily directly signalled on each programmatic change,
        this method is called by the
        :meth:`_set_alias_to_widget_value_if_sim_conf_open` method, typically
        called by the subclass implementation of this widget's main setter
        method (e.g., :meth:`QLineEdit.setText`). This method should *always* be
        called on each finalized widget change -- programmatic or otherwise.
        '''

        # logs.log_debug('In _set_alias_to_widget_value_if_sim_conf_open()...')

        # Value currently displayed by this widget.
        widget_value = self.widget_value

        # If either:
        if (
            # This widget's value remains unchanged.
            widget_value == self._widget_value_last or
            # No simulation configuration is currently open.
            not self._is_sim_open
        # Then reduce to a noop.
        ):
            return

        # Value to set this alias to, coerced from this widget's current value.
        alias_value = self._get_alias_from_widget_value()

        # Log this setting.
        logs.log_debug(
            'Setting widget "%s" alias value to %r...',
            self.obj_name, alias_value)

        # Set this alias' current value to this coerced value.
        self._sim_conf_alias.set(alias_value)

        # If this widget has a prior value to be undone...
        if self._widget_value_last is not None:
            # Push an undo command onto the stack (permitting this edit to be
            # undone) *BEFORE* updating the "_widget_value_last" variable.
            undo_cmd = QBetseeSimConfEditScalarWidgetUndoCommand(
                widget=self, value_old=self._widget_value_last)
            self._push_undo_cmd_if_safe(undo_cmd)

            # Notify all connected slots that the currently open simulation
            # configuration has received new unsaved changes *AFTER* pushing an
            # undo command onto the stack. Why? Because this method detects
            # unsaved changes by deferring to the stack state.
            self._update_sim_conf_dirty()

        # Cache this widget's current value in preparation for the next change.
        self._widget_value_last = widget_value


    def _get_alias_from_widget_value(self) -> object:
        '''
        Value displayed by this widget, coerced into a type expected by this
        simulation configuration alias.

        Design
        ----------
        The default implementation should suffice for most subclasses. However,
        subclasses for which the following test fails to validate that the value
        displayed by this widget is of the type required by this simulation
        configuration alias must override this method to do so in a
        subclass-specific manner:

            >>> isinstance(self.widget_value, self._sim_conf_alias_type)
        '''

        # Value displayed by this widget.
        alias_value = self.widget_value

        # Type(s) required by this simulation configuration alias, localized to
        # permit reduction from a tuple to non-tuple below.
        alias_type = self._sim_conf_alias_type

        # If this value is *NOT* of the type required by this simulation
        # configuration alias...
        if not isinstance(alias_value, alias_type):
            #FIXME: Non-ideal, obviously. Sadly, no better ideas come to mind.
            # If this type is an tuple of such types (e.g., "NumericSimpleTypes")
            # rather than a single type, arbitrarily coerce this value into the
            # type selected by the first item of this tuple.
            if alias_type is tuple:
                alias_type = alias_type[0]

            # Attempt to coerce this value into this type.
            alias_value = alias_type(alias_value)

        # Return this coerced value.
        return alias_value

# ....................{ SUBCLASSES                         }....................
class QBetseeSimConfEditScalarWidgetUndoCommand(QBetseeWidgetUndoCommandABC):
    '''
    Undo command generically applicable to all editable scalar simulation
    configuration widgets, implementing the application and restoration of the
    scalar contents (e.g., float, integer, string) of a single such widget.

    This subclass provides functionality specific to scalar widgets, including:

    * Automatic merging of adjacent undo commands associated with the same
      scalar widget.

    Attributes
    ----------
    _value_new : object
        New value replacing the prior value of the scalar widget associated with
        this undo command.
    _value_old : object
        Prior value of the scalar widget associated with this undo command.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        widget: QBetseeSimConfEditScalarWidgetMixin,
        value_old: object,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this undo command.

        Parameters
        ----------
        widget : QBetseeSimConfEditScalarWidgetMixin
            Scalar widget operated upon by this undo command.
        value_old : object
            Prior value of the scalar widget associated with this undo command.

        All remaining parameters are passed as is to the superclass method.
        '''

        # Initialize our superclass with all remaining arguments.
        super().__init__(
            *args,
            widget=widget,
            synopsis=widget.undo_synopsis,
            **kwargs
        )

        # Classify all passed parameters.
        self._value_old = value_old
        self._value_new = widget.widget_value

    # ..................{ SUPERCLASS ~ mandatory             }..................
    # Mandatory superclass methods required to be redefined by each subclass.

    def undo(self) -> None:

        # Defer to our superclass first.
        super().undo()

        # Undo the prior edit. To prevent infinite recursion, notify this widget
        # that an undo command is now being applied to it.
        with self._in_undo_cmd():
            self._widget.widget_value = self._value_old

            #FIXME: This focus attempt almost certainly fails across pages. If
            #this is the case, a sane general-purpose solution would be to
            #iteratively search up from the parent of this widget to the
            #eventual page of the "QStackedWidget" object containing this widget
            #and then switch to that.
            # self._widget.setFocus(Qt.OtherFocusReason)


    def redo(self) -> None:

        # Defer to our superclass first.
        super().redo()

        # Redo the prior edit. See the undo() method for further details.
        with self._in_undo_cmd():
            self._widget.widget_value = self._value_new
            # self._widget.setFocus(Qt.OtherFocusReason)

    # ..................{ SUPERCLASS ~ optional              }..................
    # Optional superclass methods permitted to be redefined by each subclass.

    def mergeWith(self, prior_undo_cmd: QUndoCommand) -> bool:
        '''
        Attempt to merge this undo command with the passed undo command
        immediately preceding this undo command on the parent undo stack,
        returning ``True`` only if this method performed this merge.

        Specifically, this method returns:

        * ``True`` if this method successfully merged both the undo and redo
          operations applied by the prior undo command into those applied by
          this undo command, in which case the prior undo command is safely
          removable from the parent undo stack.
        * ``False`` otherwise, in which case both the prior undo command and
          this undo command *must* be preserved as is the parent undo stack.

        Parameters
        ----------
        prior_undo_cmd : QUndoCommand
            Undo command immediately preceding this undo command on the parent
            undo stack.

        Returns
        ----------
        bool
            ``True`` only if these undo commands were successfully merged.
        '''

        # If this prior undo command is either of a different type *OR*
        # associated with a different widget than this undo command, these
        # commands cannot be safely merged and failure is reported.
        if not (
            self.id() == prior_undo_cmd.id() and
            self._widget == prior_undo_cmd._widget
        ):
            return False

        # Else, these commands are safely mergeable. Do so by replacing the
        # prior value of this scalar widget stored with this undo command by the
        # prior value of this scalar widget stored with this prior undo command.
        self._value_old = prior_undo_cmd._value_old

        # Report success.
        return True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Application-wide **clipboard** (i.e., platform-specific system clipboard with
which arbitrary strings may be copied and cut to and pasted from) functionality.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication
from PySide2.QtGui import QClipboard, QGuiApplication
from betse.util.type.decorator.decmemo import func_cached
from betsee.guiexception import BetseePySideClipboardException

# ....................{ EXCEPTIONS                         }....................
def die_unless_clipboard_text() -> None:
    '''
    Raise an exception unless the system clipboard's plaintext buffer is
    currently empty.

    Raises
    ----------
    BetseePySideClipboardException
        If this buffer is currently empty.

    See Also
    ----------
    :func:`is_clipboard_text`
        Further details.
    '''

    if not is_clipboard_text():
        raise BetseePySideClipboardException(
            title=QCoreApplication.translate(
                'die_if_clipboard_empty', 'Clipboard Error'),
            synopsis=QCoreApplication.translate(
                'die_if_clipboard_empty',
                'System clipboard text buffer empty.'),
        )

# ....................{ TESTERS                            }....................
def is_clipboard_text() -> bool:
    '''
    ``True`` only if the system clipboard's plaintext buffer is currently empty
    (e.g., if no plaintext has been copied or cut into the clipboard for this
    windowing session).
    '''

    return not not get_clipboard().text()

# ....................{ GETTERS                            }....................
@func_cached
def get_clipboard() -> QClipboard:
    '''
    System clipboard.

    This getter is a convenience wrapper for the stock
    :func:`QGuiApplication.clipboard` getter, reducing import statement bloat.
    '''

    return QGuiApplication.clipboard()


def get_clipboard_text() -> str:
    '''
    All text in the system clipboard's plaintext buffer if non-empty *or* raise
    an exception otherwise (e.g., if no plaintext has been copied or cut into
    the clipboard for this windowing session).

    Returns
    ----------
    str
        Most recently pasted plaintext in the system clipboard.

    Raises
    ----------
    BetseePySideClipboardException
        If this buffer is currently empty.
    '''

    # If this buffer is empty, raise an exception.
    die_unless_clipboard_text()

    # Else, this buffer is non-empty. Return its contents.
    return get_clipboard().text()



#!/usr/bin/env python3
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Top-level classes defining this application's main window.

Motivation
----------
To avoid collisions between the names assigned by PySide2's UI compiler (UIC)
to widget variables owned by the current :class:`QBetseeMainWindow` instance,
these names are typically prefixed by common substrings signifying ad-hoc
namespaces. For example, all widgets with variable names prefixed by
``sim_conf_stack_page_`` signify pages of the top-level stack widget.

For maintainability, these prefixes are centralized here rather than chaotically
dispersed throughout the codebase.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific submodules of this subpackage (i.e.,
# "betsee.gui.window"). Since those submodules must *ALWAYS* be able to safely
# import from this submodule, circularities are best avoided here.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ GLOBALS ~ sim conf                 }....................
SIM_CONF_STACK_PAGE_NAME_PREFIX = 'sim_conf_stack_page_'
'''
Substring prefixing the name of each :class:`QBetseeMainWindow` variable
providing a page of the top-level stack widget for the main window.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:class:`QRadioButton`-based simulation configuration widget subclasses.
'''

#FIXME: Submit an upstream issue requesting that:
#
#* "QButtonGroup" widgets be promotable from within Qt (Creator|Designer).
#* "QButtonGroup" widgets be treated as customary widgets accessible from the
#  widgets toolkit pane (presumably filtered into the "Containers" section).
#  Doing so addresses existing UX concerns on how "QButtonGroup" widgets are
#  currently exposed to end users from within Qt (Creator|Designer) *AND*
#  presumably automatically resolves the prior concern of promotability.

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication, Signal
from PySide2.QtWidgets import QButtonGroup, QRadioButton
# from betse.util.io.log import logs
from betse.util.type.iterable import iterables
from betse.util.type.obj import objects
from betse.util.type.types import type_check
from betsee.guiexception import BetseePySideRadioButtonException
from betsee.gui.simconf.stack.widget.abc.guisimconfwdgeditenum import (
    QBetseeSimConfEditEnumWidgetMixin)

# ....................{ SUBCLASSES                         }....................
class QBetseeSimConfEnumRadioButtonGroup(
    QBetseeSimConfEditEnumWidgetMixin, QButtonGroup):
    '''
    Simulation configuration-specific radio button group widget, permitting
    high-level enumeration members backed by low-level raw strings in external
    simulation configuration files to be interactively edited.

    Caveats
    ----------
    **Qt (Creator|Designer) provides no means of promoting :class:`QButtonGroup`
    widgets to instances of this subclass,** a longstanding deficiency with no
    short-term official solution. Instead, button groups *must* be manually
    "promoted" via the admittedly hackish
    :attr:`betsee.gui.guicache._PROMOTE_OBJ_NAME_TO_TYPE` dictionary.

    See Also
    ----------
    :class:`QBetseeSimConfComboBox`
        Alternative simulation configuration-specific widget subclass similarly
        permitting high-level enumeration members to be interactively edited.
        While more cumbersome to initialize, that subclass is preferable from
        the user experience (UX) perspective for sufficiently large enumerations
        (e.g., containing ten or more members).
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def init(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().init(*args, **kwargs)

        # For each radio button previously added to this button group...
        for radio_btn in self.buttons():
            # If this button is *NOT* a radio button, raise an exception.
            objects.die_unless_instance(obj=radio_btn, cls=QRadioButton)

            # If this button is *NOT* mapped to be the dictionary externally
            # passed to this method, raise an exception.
            if radio_btn not in self._widget_value_to_enum_member:
                raise BetseePySideRadioButtonException(
                    QCoreApplication.translate(
                        'QBetseeSimConfEnumRadioButtonGroup',
                        'Button group "{1}" radio button "{0}" not a '
                        'value of "enum_member_to_widget_value".'.format(
                            radio_btn.objectName(), self.objectName())))

    # ..................{ MIXIN ~ property : read-only       }..................
    @property
    def undo_synopsis(self) -> str:

        return QCoreApplication.translate(
            'QBetseeSimConfEnumRadioButtonGroup', 'changes to a radio button')


    @property
    def _finalize_widget_change_signal(self) -> Signal:

        # Signals defined by the "QRadioButton" class are intentionally ignored.
        # These low-level child widgets should *ONLY* ever be contained in a
        # higher-level "QButtonGroup" parent widget.
        #
        # The low-level QRadioButton.toggled() signal is emitted for the pair of
        # radio buttons in a group being enabled and disabled, thus generating
        # two emissions for each toggling of a radio button. In contrast, the
        # higher-level QButtonGroup.buttonClicked() signal returned here is
        # emitted only once for each such toggling.
        #
        # Lastly, despite nomenclature suggesting this signal to apply *ONLY* to
        # interactive clicks, Qt emits this signal on *ALL* available button
        # interactions (e.g., interactive clicks, programmatic clicks, keyboard
        # shortcut-driven clicks).
        return self.buttonClicked

    # ..................{ MIXIN ~ property : value           }..................
    @property
    def widget_value(self) -> QRadioButton:

        return self.checkedButton()


    @widget_value.setter
    @type_check
    def widget_value(self, widget_value: QRadioButton) -> None:

        # Select this radio button manually. Thanks to the mutual exclusivity of
        # radio buttons, doing so implicitly unselects all other buttons in this
        # group. For unknown reasons, the "QButtonGroup" widget is the only
        # scalar widget to provide no corresponding setCheckedButton() setter.
        widget_value.setChecked(True)


    def _reset_widget_value(self) -> None:

        # Reset this button group by arbitrarily selecting its first button.
        self.widget_value = iterables.get_item_first(self.buttons())



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for integrating :mod:`PySide2` widget classes with
XML-formatted Qt resource collection (QRC) files exported by the external Qt
Designer application.
'''

# ....................{ IMPORTS                            }....................
from betse.util.io.log import logs
from betse.util.path import files, pathnames
from betse.util.path.command import cmdrun, cmds
from betse.util.type.types import type_check

# ....................{ CONVERTERS                         }....................
@type_check
def convert_qrc_to_py_file_if_able(qrc_filename: str, py_filename: str) -> None:
    '''
    Convert the XML-formatted file with the passed ``.qrc``-suffixed filename
    *and* all binary resources referenced by this file exported by the external
    Qt Designer GUI into the :mod:`PySide2`-based Python module with the passed
    ``.py``-suffixed filename if capable of doing so *or* log a non-fatal
    warning and return otherwise.

    Dependencies
    ----------
    This function requires the optional third-party dependency ``pyside2-tools``
    distributed by The Qt Company. Specifically, this high-level function wraps
    the low-level ``pyside2-rcc`` command installed by this dependency with a
    human-usable API.

    If this command is unavailable, this function logs a non-fatal warning and
    returns *without* raising a fatal exception.

    Parameters
    ----------
    qrc_filename : str
        Absolute or relative path of the input ``.qrc``-suffixed file.
    py_filename : str
        Absolute or relative path of the output ``.py``-suffixed file.
    '''

    # Log this conversion attempt.
    logs.log_info(
        'Generating PySide2 module "%s" from "%s"...',
        pathnames.get_basename(py_filename),
        pathnames.get_basename(qrc_filename))

    # If the optional third-party dependency required by this function is
    # unavailable, log a non-fatal warning and return.
    if not cmds.is_command('pyside2-rcc'):
        logs.log_warning(
            'Skipping! Command "pyside2-rcc" not found, '
            'probably due to "pyside2-tools" not being installed.')
        return

    # If this output file is unwritable, log a non-fatal warning and return.
    if not files.is_exists_writable(py_filename):
        logs.log_warning(
            'Skipping! PySide2 module "%s" unwritable, '
            'probably due to a system-wide installation.', py_filename)
        return

    # If this input file does *NOT* exist, raise an exception.
    files.die_unless_file(qrc_filename)

    # If these files do *NOT* have the expected filetypes, raise an exception.
    pathnames.die_unless_filetype_equals(pathname=qrc_filename, filetype='qrc')
    pathnames.die_unless_filetype_equals(pathname=py_filename,  filetype='py')

    # Convert this input file to this output file or raise an exception if
    # unsuccessful. For debuggability, this command's stdout and stderr is
    # redirected to this application's stdout, stderr, and logging file handles.
    cmdrun.log_output_or_die(
        command_words=('pyside2-rcc', '-o', py_filename, qrc_filename))

    #FIXME: The contents of this output "py_filename" should additionally be
    #opened for writing and prefixed by a shebang line running the active Python
    #interpreter. See the "guiui" submodule for relevant logic.



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Collection of the absolute paths of numerous critical files and directories
describing the structure of this application on the local filesystem.

These are intended for consumption by both this application and downstream
reverse dependencies of this application. For portability, these paths are
initialized in a system-aware manner guaranteed to be sane under insane
installation environments -- including PyInstaller-frozen executables and
:mod:`setuptools`-installed script wrappers.

See Also
----------
:mod:`betsee.util.path.guipathsys`
    Collection of the absolute paths of numerous critical files and directories
    describing the structure of the local filesystem.
'''

#FIXME: The current globals-based approach is inefficient in the case of BETSE
#being installed as a compressed EGG rather than an uncompressed directory. In
#the former case, the current approach (namely, the call to
#resources.get_pathname() performed below) silently extracts the entirety of
#this egg to a temporary setuptools-specific cache directory. That's bad. To
#circumvent this, we'll need to refactor the codebase to directly require only
#"file"-like objects rather than indirectly requiring the absolute paths of
#data resources that are then opened as "file"-like objects.
#
#Specifically, whenever we require a "file"-like object for a codebase resource,
#we'll need to call the setuptools-specific pkg_resources.resource_stream()
#function rather than attempting to open the path given by a global below.
#Ultimately, *ALL* of the codebase-specific globals declared below (e.g.,
#"DATA_DIRNAME") should go away.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory dependencies,
# the top-level of this module may import *ONLY* from packages guaranteed to
# exist at installation time (i.e., standard Python packages). Likewise, to
# avoid circular import dependencies, the top-level of this module should avoid
# importing application packages except where explicitly required.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse import pathtree as betse_pathtree
from betse.util.path import dirs, files, pathnames
from betse.util.type.decorator.decmemo import func_cached
from betsee import guimetadata

# ....................{ GETTERS ~ dir : data               }....................
@func_cached
def get_data_dirname() -> str:
    '''
    Absolute path of this application's top-level data directory if found *or*
    raise an exception otherwise (i.e., if this directory is *not* found).

    This directory contains application-internal resources (e.g., media files)
    required at application runtime.
    '''

    # Avoid circular import dependencies.
    import betsee

    # Absolute path of this directory.
    data_dirname = pathnames.get_app_pathname(package=betsee, pathname='data')

    # If this directory is not found, raise an exception.
    dirs.die_unless_dir(data_dirname)

    # Return the absolute path of this directory.
    return data_dirname


@func_cached
def get_data_py_dirname() -> str:
    '''
    Absolute path of this application's data subdirectory containing
    pure-Python modules and packages generated at runtime by this application if
    found *or* raise an exception otherwise (i.e., if this directory is *not*
    found).
    '''

    # Create this directory if needed and return its dirname.
    return dirs.join_and_die_unless_dir(get_data_dirname(), 'py')


@func_cached
def get_data_qrc_dirname() -> str:
    '''
    Absolute path of this application's data subdirectory containing
    XML-formatted Qt resource collection (QRC) files exported by the external Qt
    Designer application and all binary resource files listed in these files if
    found *or* raise an exception otherwise (i.e., if this directory is *not*
    found).
    '''

    # Return this dirname if this directory exists or raise an exception.
    return dirs.join_and_die_unless_dir(get_data_dirname(), 'qrc')


@func_cached
def get_data_ui_dirname() -> str:
    '''
    Absolute path of this application's data subdirectory containing
    XML-formatted user interface (UI) files exported by the external Qt Designer
    application if found *or* raise an exception otherwise (i.e., if this
    directory is *not* found).
    '''

    # Return this dirname if this directory exists or raise an exception.
    return dirs.join_and_die_unless_dir(get_data_dirname(), 'ui')

# ....................{ GETTERS ~ dir : dot                }....................
@func_cached
def get_dot_dirname() -> str:
    '''
    Absolute path of this application's top-level dot directory in the home
    directory of the current user, silently creating this directory if *not*
    already found.

    This directory contains user-specific files (e.g., generated Python modules)
    both read from and written to at application runtime. These are typically
    plaintext files consumable by external users and third-party utilities.

    For tidiness, this directory currently resides under BETSE's dot directory
    (e.g., ``~/.betse/betsee`` under Linux).
    '''

    # Create this directory if needed and return its dirname.
    return dirs.join_and_make_unless_dir(
        betse_pathtree.get_dot_dirname(), guimetadata.SCRIPT_BASENAME)

# ....................{ GETTERS ~ file : data              }....................
@func_cached
def get_data_qrc_filename() -> str:
    '''
    Absolute path of the XML-formatted Qt resource collection (QRC) file
    exported by the external Qt Designer application structuring all external
    resources (e.g., icons) required by this application's main window if found
    *or* raise an exception otherwise (i.e., if this file is *not* found).
    '''

    # Return this filename if this file exists or raise an exception.
    #
    # Note that this basename *MUST* be the same as that specified by the
    # "resources" attribute of all XML tags contained in the file whose path is
    # given by the get_data_ui_filename() function. Why? Because obfuscatory Qt.
    return files.join_and_die_unless_file(
        get_data_qrc_dirname(), guimetadata.SCRIPT_BASENAME + '.qrc')


@func_cached
def get_data_ui_filename() -> str:
    '''
    Absolute path of the XML-formatted user interface (UI) file exported by the
    external Qt Designer application structuring this application's main window
    if found *or* raise an exception otherwise (i.e., if this file is *not*
    found).
    '''

    # Return this filename if this file exists or raise an exception.
    return files.join_and_die_unless_file(
        get_data_ui_dirname(), guimetadata.SCRIPT_BASENAME + '.ui')

# ....................{ GETTERS ~ file : data              }....................
@func_cached
def get_data_py_qrc_filename() -> str:
    '''
    Absolute path of the pure-Python module generated from the XML-formatted Qt
    resource collection (QRC) file exported by the external Qt Designer
    application structuring all external resources (e.g., icons) required by
    this application's main window if found *or* raise an exception otherwise
    (i.e., if this directory is *not* found).

    Caveats
    ----------
    This module is guaranteed to be importable but *not* necessarily be
    up-to-date with the input paths from which this module is dynamically
    regenerated at runtime. The caller is assumed to do so explicitly.

    See Also
    ----------
    :mod:`betsee.gui.guicache`
        Submodule dynamically generating this module.
    '''

    # Note that this basename *MUST* be:
    #
    # * Prefixed by the same basename excluding filetype returned by the
    #   get_data_qrc_filename() function.
    # * Suffixed by "_rc.py". Why? Because the Python code generated at runtime
    #   by the "pyside2uic" package assumes this to be the case. Naturally, this
    #   assumption is *NOT* configurable.
    return files.join_and_die_unless_file(
        get_data_py_dirname(), guimetadata.MAIN_WINDOW_QRC_MODULE_NAME + '.py')


@func_cached
def get_data_py_ui_filename() -> str:
    '''
    Absolute path of the pure-Python module generated from the XML-formatted
    user interface (UI) file exported by the external Qt Designer application
    structuring this application's main window if found *or* raise an exception
    otherwise (i.e., if this directory is *not* found).

    Caveats
    ----------
    This module is guaranteed to be importable but *not* necessarily be
    up-to-date with the input paths from which this module is dynamically
    regenerated at runtime. The caller is assumed to do so explicitly.

    See Also
    ----------
    :mod:`betsee.gui.guicache`
        Submodule dynamically generating this module.
    '''

    return files.join_and_die_unless_file(
        get_data_py_dirname(), guimetadata.MAIN_WINDOW_UI_MODULE_NAME + '.py')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **non-pooled worker** (i.e., thread-safe object implementing
generically startable, pausable, resumable, and haltable business logic in a
multithreaded manner intended to be moved to the thread encapsulated by a
:class:`QThread` object) classes.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QObject, Signal, Slot  # QCoreApplication,
from betse.exceptions import BetseMethodUnimplementedException
from betse.util.io.log import logs
# from betse.util.type.types import type_check
from betsee.guiexception import BetseePySideThreadWorkerStopException
from betsee.util.thread import guithread
from betsee.util.thread.guithreadenum import ThreadWorkerState
from betsee.util.widget.abc.guiwdgabc import QBetseeObjectMixin

# ....................{ SUPERCLASSES                       }....................
class QBetseeLoneThreadWorkerABC(QBetseeObjectMixin, QObject):
    '''
    Abstract base class of all low-level **non-pooled worker** (i.e.,
    thread-safe object implementing generically startable, pausable, resumable,
    and haltable business logic in a multithreaded manner intended to be adopted
    by the thread encapsulated by a :class:`QBetseeLoneThread` object)
    subclasses.

    By default, workers are recyclable and hence may be reused (i.e., have their
    :meth:`start` slots signalled) an arbitrary number of times within any
    arbitrary thread. Where undesirable, see the
    :cless:`QBetseeLoneThreadWorkerThrowawayABC` for an alternative superclass
    constraining workers to be non-recyclable.

    Caveats
    ----------
    This obsolete superclass has been superceded by the superior
    :class:`betse.util.thread.pool.guipoolwork.QBetseeThreadPoolWorker`
    superclass, whose :class:`QRunnable`-based API requires substantially less
    boilerplate.

    Attributes
    ----------
    _state : ThreadWorkerState
        Current execution state of this worker. For thread-safety, this state
        should *not* be externally accessed by objects residing in other
        threads. Doing so safely would require thread-safe mutual exclusion
        (e.g., with a dedicated :class:`QMutexLocker` context manager), which
        currently exceeds the mandate of this superclass.

    See Also
    ----------
    https://codereview.stackexchange.com/a/173258/124625
        StackOverflow answer strongly inspiring this implementation.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:
        '''
        Initialize this worker.

        Caveats
        ----------
        This method intentionally accepts *no* passed parameters and hence
        cannot be passed a parent `QObject`. So, this worker is **unparented**
        (i.e., has no such parent). Why? Because this worker will be
        subsequently adopted into a different thread than the original thread in
        which this worker was instantiated. However, most candidate `QObject`
        parents of this worker would presumably reside in that original thread.
        Objects in different threads should typically *not* control the
        lifecycle of each other, as the parent of a child `QObject` does;
        doing so typically violates thread-safety. (That's bad.)
        '''

        # Initialize our superclass with *NO* passed parameters. (See above.)
        super().__init__()

        # Default this worker's initial state to the idle state.
        self._state = ThreadWorkerState.IDLE

        # Connect this worker's external-facing signals to corresponding slots.
        self.start_signal .connect(self.start)
        self.stop_signal  .connect(self.stop)
        self.pause_signal .connect(self.pause)
        self.resume_signal.connect(self.resume)

    # ..................{ SIGNALS ~ external                 }..................
    # Signals externally emitted by callers owning instances of this superclass.

    #FIXME: Obviously, requiring a string be unconditionally passed to all
    #start_signal() invocations is... well, insane. Instead, this should be
    #refactored as follows:
    #
    #* Define a new "QBetseeLoneThreadWorkerStartArgsABC(QObject)" class in this
    #  or possibly another submodule (e.g., "guiworkerstartcls").
    #* Rewrite this signal to simply read:
    #     start_signal = Signal(QBetseeLoneThreadWorkerStartArgsABC)
    #* Rewrite the start() slot and _work() methods defined below to similarly
    #  accept instances of this "QBetseeLoneThreadWorkerStartArgsABC" class rather
    #  than... raw strings. *sigh*
    #* Refactor all subclasses of this class similarly.
    #* Refactor all emissions of this signal to emit
    #  "QBetseeLoneThreadWorkerStartArgsABC" instances rather than raw strings.
    #FIXME: Actually, we didn't realize that signals could emit arbitrary Python
    #objects. Apparently, they can. Ergo, the above plan should be refactored to
    #instead subclass "ThreadWorkerStartArgsABC(object, meta=ABCMeta)". Maybe?
    start_signal = Signal(str)
    '''
    Signal connected the :meth:`start` slot starting this worker.

    This signal is a caller convenience simplifying worker usage. Alternately,
    callers may elect to connect caller-defined signals to this slot as needed.
    '''


    stop_signal = Signal()
    '''
    Signal connected the :meth:`stop` slot starting this worker.

    This signal is a caller convenience simplifying worker usage. Alternately,
    callers may elect to connect caller-defined signals to this slot as needed.
    '''


    pause_signal = Signal()
    '''
    Signal connected the :meth:`pause` slot pausing this worker.

    This signal is a caller convenience simplifying worker usage. Alternately,
    callers may elect to connect caller-defined signals to this slot as needed.
    '''


    resume_signal = Signal()
    '''
    Signal connected the :meth:`resume` slot resuming this worker.

    This signal is a caller convenience simplifying worker usage. Alternately,
    callers may elect to connect caller-defined signals to this slot as needed.
    '''

    # ..................{ SIGNALS ~ internal                 }..................
    # Signals internally emitted by instances of this superclass.

    started = Signal()
    '''
    Signal emitted immediately on entering the :meth:`start` slot starting this
    worker.

    This signal is emitted *before* that slot performs *any* subclass-specific
    business logic for this worker.
    '''


    finished = Signal(bool)
    '''
    Signal emitted immediately before returning from the :meth:`start` slot for
    this worker, passed either ``True`` if this slot successfully performed all
    subclass-specific business logic for this worker (i.e., if the :meth:`_work`
    method returns *without* raising exceptions) *or* ``False`` otherwise.

    This signal is emitted *after* that slot performs *all* subclass-specific
    business logic for this worker.
    '''

    # ..................{ SLOTS                              }..................
    #FIXME: Obviously, forcing "str" usage is awful. See above for solutions.
    @Slot(str)
    def start(self, arbitrary_str: str) -> None:
        '''
        Slot performing *all* subclass-specific business logic for this worker.

        This slot works in a thread-safe manner safely pausable and stoppable at
        any time (e.g., by emitting a signal connected to the :meth:`pause` or
        :meth:`stop` slots).

        States
        ----------
        If this worker is in the :attr:`ThreadWorkerState.IDLE` state, this
        slot changes to the :attr:`ThreadWorkerState.WORKING` state and calls
        the subclass :meth:`_work` method.

        If this worker is in the :attr:`ThreadWorkerState.PAUSED` state, this
        slot interprets this signal as a request to resume the work presumably
        previously performed by this worker by a prior signalling of this slot.
        To avoid reentrancy issues, this slot changes to the
        :attr:`ThreadWorkerState.WORKING` state and immediately returns.
        Assuming that a prior call to this slot is still executing, that call
        will internally detect this change and resume working as expected.

        If this worker is in the :attr:`ThreadWorkerState.WORKING` state, this
        slot interprets this signal as an accidental attempt by an external
        caller to re-perform the work concurrently being performed by a prior
        call to this slot. In that case, this slot safely logs a non-fatal
        warning and immediately returns.

        See the :meth:`pause` slot for commentary on these design decisions.

        Signals
        ----------
        This slot emits the following signals:

        * :attr:`started` immediately *before* this slot performs any
          subclass-specific business logic for this worker.
        * :attr:`finished` immediately *after* this slot performs all
          subclass-specific business logic for this worker.

        Caveats
        ----------
        Subclasses must override the :meth:`_work` method rather than this slot
        to perform subclass-specific business logic. This slot is neither
        intended nor designed to be overriden by subclasses.
        '''

        # True only if the _work() method called below returns successfully
        # (i.e., *WITHOUT* raising exceptions). Defaults to False for safety.
        is_success = False

        # If this worker is currently paused, resume the prior call to this
        # start() slot by changing to the working state and returning. See the
        # method docstring for commentary.
        if self._state is ThreadWorkerState.PAUSED:
            logs.log_debug(
                'Resuming thread "%d" worker "%s" '
                'via reentrant start() slot...',
                guithread.get_thread_current_id(), self.obj_name)
            self._state = ThreadWorkerState.WORKING
            return

        # If this worker is currently running, resume the prior call to this
        # start() slot by preserving the working state and returning. See the
        # method docstring for commentary. Unlike the prior logic, this logic
        # constitutes a non-fatal error and is logged as such.
        if self._state is ThreadWorkerState.WORKING:
            logs.log_warning(
                'Ignoring attempt to reenter '
                'thread "%d" worker "%s" start() slot.',
                guithread.get_thread_current_id(), self.obj_name)
            return

        # Log this beginning.
        logs.log_debug(
            'Starting thread "%d" worker "%s"...',
            guithread.get_thread_current_id(), self.obj_name)

        # Change to the working state.
        self._state = ThreadWorkerState.WORKING

        # Notify external subscribers *BEFORE* beginning subclass work.
        self.started.emit()

        # Attempt to...
        try:
            # If this worker or this worker's thread has been externally
            # requested to stop immediately after being requested to start,
            # promptly respect this wish on behalf of all workers. Although an
            # unlikely edge case, the fixed cost of this test is negligible.
            self._halt_work_if_requested()

            # Perform all subclass work.
            self._work(arbitrary_str)

            # Note the _work() method called above to have returned successfully
            # (i.e., *WITHOUT* raising exceptions).
            is_success = True

            # If the state of this worker is still the working state, set this
            # state to the idle (i.e., non-working) state to preserve sanity.
            if self._state is ThreadWorkerState.WORKING:
                self._state = ThreadWorkerState.IDLE
        # If a periodic call to the _halt_work_if_requested() method performed
        # within the above call detects either this worker or this worker's
        # thread has been externally requested to stop, do so gracefully by...
        # doing absolutely nothing. Welp, that was easy.
        except BetseePySideThreadWorkerStopException:
            pass

        # Log this completion.
        logs.log_debug(
            'Finishing thread "%d" worker "%s" with success status "%r"...',
            guithread.get_thread_current_id(),
            self.obj_name,
            is_success)

        # Notify external subscribers of whether this work succeeded or not
        # *AFTER* completing all such work.
        self.finished.emit(is_success)


    @Slot()
    def stop(self) -> None:
        '''
        Slot stopping all work performed by this worker.

        This slot prematurely halts this work in a thread-safe manner. Whether
        this work is safely restartable (e.g., by emitting a signal connected to
        the :meth:`start` slot) is a subclass-specific implementation detail.
        Subclasses may voluntarily elect to either prohibit or permit restarts.

        States
        ----------
        If this worker is in the :attr:`ThreadWorkerState.IDLE` state, this
        slot silently reduces to a noop and preserves the existing state. In
        this case, this worker remains idle.

        If this worker is in either the :attr:`ThreadWorkerState.WORKING` or
        :attr:`ThreadWorkerState.PAUSED`, implying this worker to either
        currently be or recently have been working, this slot changes the
        current state to the :attr:`ThreadWorkerState.IDLE` state. In either
        case, this worker ceases working.
        '''

        # Log this change.
        logs.log_debug(
            'Stopping thread "%d" worker "%s"...',
            guithread.get_thread_current_id(), self.obj_name)

        # If this worker is currently working or paused, stop this worker.
        if self._state is not ThreadWorkerState.IDLE:
            self._state = ThreadWorkerState.IDLE

    # ..................{ SLOTS ~ pause                      }..................
    @Slot()
    def pause(self) -> None:
        '''
        Slot pausing all work performed by this worker.

        This slot temporarily halts this work in a thread-safe manner safely
        resumable at any time (e.g., by emitting a signal connected to the
        :meth:`resume` slot).

        States
        ----------
        If this worker is *not* currently working, this slot silently reduces to
        a noop. While raising a fatal exception in this edge case might
        superficially appear to be reasonable, the queued nature of signal-slot
        connections introduces unavoidable delays in event delivery and hence
        slot execution. In particular, raising an exception would introduce a
        race condition between the time that a user interactively requests a
        working worker to be paused and that worker's completion of its work.
        '''

        # If this worker is *NOT* currently working...
        if self._state is not ThreadWorkerState.WORKING:
            # Log this attempt.
            logs.log_debug(
                'Ignoring attempt to pause idle or already paused '
                'thread "%d" worker "%s".',
                guithread.get_thread_current_id(), self.obj_name)

            # Safely reduce to a noop.
            return
        # Else, this worker is currently working.

        # Log this change.
        logs.log_debug(
            'Pausing thread "%d" worker "%s"...',
            guithread.get_thread_current_id(), self.obj_name)

        # Change this worker's state to paused.
        self._state = ThreadWorkerState.PAUSED

        # While this worker's state is paused, wait for the resume() slot to be
        # externally signalled by another object (typically in another thread).
        while self._state is ThreadWorkerState.PAUSED:
            guithread.wait_for_thread_current_event_if_none()


    @Slot()
    def resume(self) -> None:
        '''
        Slot **resuming** (i.e., unpausing) this worker.

        This slot resumes working in a thread-safe manner safely re-pausable at
        any time (e.g., by re-emitting a signal connected to the :meth:`pause`
        slot).

        States
        ----------
        If this worker is in either of the following states, this slot silently
        reduces to a noop and preserves the existing state:

        * :attr:`ThreadWorkerState.IDLE`, implying this worker to *not*
          currently be paused. In this case, this worker remains idle.
        * :attr:`ThreadWorkerState.WORKING`, implying this worker to already
          have been resumed. In this case, this worker remains working.

        See the :meth:`pause` slot for commentary on this design decision.
        '''

        # Log this change.
        logs.log_debug(
            'Resuming thread "%d" worker "%s"...',
            guithread.get_thread_current_id(), self.obj_name)

        # If this worker is currently paused, unpause this worker.
        if self._state is ThreadWorkerState.PAUSED:
            self._state = ThreadWorkerState.WORKING

    # ..................{ METHODS ~ abstract                 }..................
    # Abstract methods required to be redefined by subclasses.

    #FIXME: Obviously, forcing "str" usage is awful. See above for solutions.
    def _work(self, arbitrary_str: str) -> None:
        '''
        Perform *all* subclass-specific business logic for this worker.

        The superclass :meth:`start` slot internally calls this method in a
        thread-safe manner safely pausable *and* stoppable at any time (e.g., by
        emitting a signal connected to the :meth:`pause` or :meth:`stop` slots).

        Design
        ----------
        Subclasses are required to redefine this method to perform this logic in
        an iterative manner periodically calling the
        :meth:`_halt_work_if_requested` method.

        If either:

        * This worker has been externally signalled to stop (e.g., by emitting a
          signal connected to the :meth:`stop` slot).
        * The thread of execution currently running this worker has been
          externally requested to stop (e.g., by calling the
          :func:`guithread.halt_thread_work` function).

        Then the next such call to the :meth:`_halt_work_if_requested` method
        will raise an exception caught by the parent :meth:`start` slot,
        signalling that slot to immediately terminate this worker. Ergo, that
        method should be called *only* when the subclass is in an
        **interruptible state** (i.e., a self-consistent internal state in which
        this subclass is fully prepared to be immediately terminated).
        '''

        # The next best thing to a properly abstract method, given "QObject"
        # constraints against declaring an "ABCMeta" metaclass. *shrug*
        raise BetseMethodUnimplementedException()

    # ..................{ METHODS ~ concrete                 }..................
    # Concrete methods intended to be called but *NOT* overriden by subclasses.

    def _halt_work_if_requested(self) -> None:
        '''
        Raise an exception if either:

        * This worker has been externally signalled to stop (e.g., by emitting a
          signal connected to the :meth:`stop` slot).
        * The thread of execution currently running this worker has been
          externally requested to stop (e.g., by calling the
          :func:`guithread.halt_thread_work` function).

        This function is intended to be periodically called by the subclass
        :meth:`_work` function. The exception raised by this function is
        guaranteed to be caught by the :meth:`start` method calling that
        :meth:`_work` function.

        Caveats
        ----------
        This function imposes minor computational overhead and hence should be
        called intermittently (rather than overly frequently). Notably, each
        call to this method processes *all* pending events currently queued with
        this worker's thread -- including those queued for all other workers
        currently running in this thread.

        Raises
        ----------
        BetseePySideThreadWorkerStopException
            If this worker or this worker's thread of execution has been
            signalled or requested to be stopped.
        '''

        # Process all pending events currently queued with this worker's thread,
        # notably including any incoming signalling of the stop() slot by
        # objects in other threads..
        guithread.process_thread_current_events()

        # If either:
        if (
            # This worker has been externally signalled to stop...
            self._state is ThreadWorkerState.IDLE or
            # This worker's thread has been externally requested to stop...
            guithread.should_halt_thread_work()
        # Then
        ):
            # Log this interrupt.
            logs.log_debug(
                'Interrupting thread "%d" worker "%s"...',
                guithread.get_thread_current_id(), self.obj_name)

            # Instruct the parent start() slot to stop.
            raise BetseePySideThreadWorkerStopException('So say we all.')

# ....................{ SUPERCLASSES ~ throwaway           }....................
class QBetseeLoneThreadWorkerThrowawayABC(QBetseeLoneThreadWorkerABC):
    '''
    Abstract base class of all low-level **throw-away non-pooled worker** (i.e.,
    non-pooled worker guaranteed to be garbage-collected after completing its
    work) subclasses.

    Equivalently, this is the superclass of all single-use, one-shot, one-time,
    non-recyclable non-pooled workers. By default, workers are recyclable and
    hence may be reused (i.e., have their :meth:`start` slots signalled) an
    arbitrary number of times within any arbitrary thread. Instances of this
    superclass are non-recyclable and hence may *not* be reused more than once,
    however.

    Usage
    ----------
    Ideally, the :meth:`start` slot of each instance of this superclass should
    be signalled at most once. Immediately *after* performing all
    subclass-specific work via the :meth:`_work` method, this slot (in order):

    #. Emits the :attr:`finished` signal.
    #. Signals itself to be garbage-collected via the :meth:`deleteLater` slot.
    #. Returns.

    After returning from the :meth:`start` slot, all external references to this
    worker are effectively invalidated and hence equivalent to ``None``; no such
    worker may be safely used for *any* subsequent purposes.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        #FIXME: Sadly, this connection now appears to induce segmentation
        #faults on worker completion. Our intuitive suspicion is that the
        #"finished" signal now emits booleans, which are probably being passed
        #to the deleteLater() slot, which doesn't expect to receive boolean
        #parameters and hence... fatally crashes? It seems to be the case,
        #curiously. The solution, of course, would be to connect the "finished"
        #signal to a new slot accepting (and ignoring) a passed boolean and
        #simply calling the self.deleteLater() slot directly -- say:
        #
        #   @Slot(bool)
        #   def _schedule_deletion(self, is_success: bool) -> None:
        #       '''
        #       Schedule this thread worker and *all* :class:`QObject` instances
        #       owned by this worker for subsequent deletion.
        #       '''
        #
        #       self.deleteLater()

        # Garbage collect all child objects of this parent worker *AFTER* this
        # worker gracefully (i.e., successfully) terminates.
        # self.finished.connect(self.deleteLater)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for integrating :mod:`PySide2` widget classes
with XML-formatted user interface (UI) files exported by the external Qt
Designer application.
'''

# ....................{ IMPORTS                           }....................
from PySide2 import QtWidgets
from PySide2.QtCore import QCoreApplication, QObject
from betse.util.io import iofiles
from betse.util.io.log import logs
from betse.util.path import files, pathnames
from betse.util.py import pyident, pymodule, pys
from betse.util.py.pyident import IDENTIFIER_UNQUALIFIED_REGEX
from betse.util.type import types
from betse.util.type.cls import classes
from betse.util.type.text import regexes
from betse.util.type.types import type_check, MappingType, SequenceTypes
from betsee import guimetadata
from betsee.guiexception import BetseeCacheException
from betsee.lib import guilib
from io import StringIO

# ....................{ GLOBALS                           }....................
# To avoid conflict with PySide2-generated attribute names, obfuscate this
# global's name in an application-specific manner.
BASE_CLASSES_GLOBAL_NAME = '__{}_BASE_CLASSES'.format(guimetadata.NAME)
'''
Unqualified name of the application-specific global appended to the Python code
generated by the :func:`convert_ui_to_py_file` function.

This global declares the sequence of all base classes that the main window Qt
widget subclass for this application is expected to subclass (in order).
'''

# ....................{ GETTERS                           }....................
@type_check
def get_ui_module_base_classes(ui_module_name: str) -> SequenceTypes:
    '''
    Sequence of all base classes declared by the user interface (UI) module
    with the passed name and intended to be subclassed by the Qt widget
    subclass implementing this UI.

    The :func:`convert_ui_to_py_file` function is assumed to have generated
    this module from an XML-formatted file exported by the external Qt Designer
    GUI.

    Parameters
    ----------
    ui_module_name : str
        Fully-qualified name of the UI module to be imported and inspected by
        this function. Since this module is dynamically generated at runtime
        and hence *not* guaranteed to exist, this function explicitly validates
        this module to both exist and declare this sequence..

    Raises
    ----------
    BetseModuleException
        If this module is unimportable.
    BetseeCacheException
        If this module declares no such sequence.
    '''

    # UI module if importable *OR* raise an exception otherwise.
    ui_module = pymodule.import_module(
        module_name=ui_module_name,
        exception_message=QCoreApplication.translate(
            'get_ui_module_base_classes',
            'Module "{0}" not found (e.g., as'
            'module "betsee.gui.window.guimainwindow" imported before '
            'module "betsee.gui.guicache").'.format(ui_module_name)
        ))

    # Sequence of all base classes if declared by this module or None otherwise.
    ui_base_classes = getattr(ui_module, BASE_CLASSES_GLOBAL_NAME, None)

    # Fully-qualified name of this sequence for use in exception messages.
    ui_base_classes_name = '{}.{}'.format(
        ui_module_name, BASE_CLASSES_GLOBAL_NAME)

    # If no such sequence is declared, raise an exception.
    if ui_base_classes is None:
        raise BetseeCacheException(QCoreApplication.translate(
            'get_ui_module_base_classes',
            'Sequence "{0}" undefined.'.format(ui_base_classes_name)))

    # If this sequence is *NOT* a sequence, raise an exception. While this
    # check is also performed by the @type_check decorator, the following
    # length check implicitly assumes this variable to be a sequence.
    if not types.is_sequence_nonstr(ui_base_classes):
        raise BetseeCacheException(QCoreApplication.translate(
            'get_ui_module_base_classes',
            'Variable "{0}" type {1!r} not a non-string sequence.'.format(
                ui_base_classes_name, type(ui_base_classes))))

    # If this sequence does *NOT* contain two items, raise an exception.
    if len(ui_base_classes) != 2:
        raise BetseeCacheException(QCoreApplication.translate(
            'get_ui_module_base_classes',
            'Sequence "{0}" length {1} not 2.'.format(
                ui_base_classes_name, len(ui_base_classes))))

    # Else, return this sequence.
    return ui_base_classes

# ....................{ CONVERTERS                        }....................
@type_check
def convert_ui_to_py_file_if_able(
    ui_filename: str,
    py_filename: str,
    promote_obj_name_to_class: MappingType,
) -> None:
    '''
    Convert the XML-formatted file with the passed ``.ui``-suffixed filename
    exported by the external Qt Designer GUI into the :mod:`PySide2`-based
    Python module with the passed ``.py``-suffixed filename if capable of doing
    so *or* log a non-fatal warning and return otherwise.

    Dependencies
    ----------
    This function requires the optional third-party dependency
    ``pyside2-tools`` distributed by The Qt Company. Specifically, this
    high-level function wraps the low-level
    :meth:`pyside2uic.Compiler.compiler.UICompiler.compileUi` method installed
    by this dependency with a human-usable API.

    If the :mod:`pyside2uic` package is unimportable, this function logs a
    non-fatal warning and returns *without* raising a fatal exception.

    Design
    ----------
    This function silently overwrites this output file with the dynamically
    compiled contents of a pure-Python model declaring:

    * A custom helper class defining the following methods:

      * ``setupUi()``, accepting an instance of the :mod:`PySide2` widget base
        class listed as the last item of the global sequence described below
        and customizing this widget as specified by this input UI file.
      * ``retranslateUi()``, accepting a similar instance and translating all
        strings displayed by this widget from their default language
        (typically, English) into the locale of the current user.

    * A global sequence variable whose:

      * Name is the value of the
        :data:`BASE_CLASSES_GLOBAL_NAME` global string variable,
      * Value is the sequence of all base classes that the main Qt window
        widget class for this application *must* subclass (in order):

        #. The custom helper class described above.
        #. The :mod:`PySide2` widget base class of the object to be passed to
           all methods defined by this class (e.g., ``setupUi()``). These
           methods do *not* validate this to be the case but do raise
           non-human-readable exceptions when passed objects that are *not*
           instances of this base class. This base class is defined by this
           input UI file and is typically one of the following:

           * :class:`PySide2.QWidget.QDialog`.
           * :class:`PySide2.QWidget.QWidget`.
           * :class:`PySide2.QWidget.QMainWindow`.

    This function provides a :mod:`PySide2`-specific implementation of the
    PyQt-specific :func:`PyQt5.uic.loadUiType` function, albeit with a more
    typesafe API. For unknown reasons (presumably laziness), :mod:`PySide2`
    lacks an analogue to this function.

    For equally unknown reasons, the class returned by that function is
    commonly referred to as the "form class" in most Qt documentation. This
    class has no relation to the :class:`PySide2.QLayout.QFormLayout` base
    class and is thus arguably *not* a "form class." For disambiguity, we
    prefer to refer to this class as simply the "generated UI class."

    Parameters
    ----------
    ui_filename : str
        Absolute or relative path of the input ``.ui``-suffixed file.
    py_filename : str
        Absolute or relative path of the output ``.py``-suffixed file.
    promote_obj_name_to_class : MappingType
        Dictionary mapping from the name of each instance variable of the main
        window to manually reinstantiate to the application-specific widget
        subclass to declare that variable to be an instance of. This dictionary
        facilitates the manual "promotion" of widgets for which the Qt
        (Creator|Designer) GUI currently provides no means of official
        promotion, notably including :class:`QButtonGroup` widgets.

    Returns
    ----------
    type
        UI class generated from the XML-formatted file with this filename
        exported by the external Qt Designer application.

    See Also
    ----------
    http://pyqt.sourceforge.net/Docs/PyQt5/designer.html
        :mod:`PyQt5`-specific documentation detailing this conversion. Although
        :mod:`PyQt5`-specific, this documentation probably serves as the
        canonical resource for understanding this function's internal
        behaviour.
    '''

    # Log this conversion attempt.
    logs.log_info(
        'Generating PySide2 module "%s" from "%s"...',
        pathnames.get_basename(py_filename),
        pathnames.get_basename(ui_filename))

    # If the optional third-party dependency required by this function is
    # unavailable, log a non-fatal warning and return.
    if not guilib.is_runtime_optional('pyside2uic'):
        logs.log_warning(
            'Skipping! Package "pyside2uic" unimportable, '
            'probably due to "pyside2-tools" not being installed.')
        return

    # If this output file is unwritable, log a non-fatal warning and return.
    if not files.is_exists_writable(py_filename):
        logs.log_warning(
            'Skipping! PySide2 module "%s" unwritable, '
            'probably due to a system-wide installation.', py_filename)
        return

    # Optional third-party dependencies required by this function.
    pyside2uic = guilib.import_runtime_optional('pyside2uic')
    UICompiler = pyside2uic.Compiler.compiler.UICompiler

    # If this input file does *NOT* exist, raise an exception.
    files.die_unless_file(ui_filename)

    # If these files do *NOT* have the expected filetypes, raise an exception.
    pathnames.die_unless_filetype_equals(pathname=ui_filename, filetype='ui')
    pathnames.die_unless_filetype_equals(pathname=py_filename, filetype='py')

    # File-like string buffer containing the Python code converted from this
    # file to be subsequently evaluated.
    ui_code_str_buffer = StringIO()

    # Dictionary of high-level metadata describing the high-level types
    # produced by converting this file into this string buffer, containing the
    # following key strings:
    #
    # * "baseclass", whose value is the name of the PySide2-specific widget
    #   base class that objects passed to the setupUi() and retranslateUi()
    #   methods of the generated UI class are required to be instances of.
    # * "uiclass", whose value is the name of the generated UI class defining
    #   the setupUi() and retranslateUi() methods.
    # * "widgetname", whose value is... an unknown string. (Not our fault.)
    #
    # See the UICompiler.compileUi() method implementation for details.
    #
    # Note that most online examples call the slightly higher-level
    # pyside2uic.compileUi() function internally calling the lower-level
    # UICompiler.compileUi() method in the exact same manner as below. The
    # former has significant disadvantages and is thus ignored in favour of the
    # latter. In particular, the pyside2uic.compileUi() function:
    #
    # * Does *NOT* return the dictionary returned by the UICompiler.compileUi()
    #   method required below. While the contents of this dictionary are
    #   technically reverse engineerable by manually parsing the XML of this
    #   ".ui" file for the corresponding elements (e.g., via the
    #   "xml.etree.ElementTree" package), doing so needlessly incurs space,
    #   time, and code complexity costs. Due presumably to path dependency,
    #   most online examples (insanely) do so.
    # * Adds no meaningful advantages over the UICompiler.compileUi() method
    #   for most common cases, including this case.
    #
    # In short, the pyside2uic.compileUi() function is useless and no one
    # should ever call it.
    ui_code_metadata = UICompiler().compileUi(
        input_stream=ui_filename,
        output_stream=ui_code_str_buffer,

        # Force all generated imports to be absolute rather than relative.
        from_imports=False,
    )

    # Name of the custom class to be generated.
    ui_form_class_name = ui_code_metadata['uiclass']

    # Name of the PySide2 base class expected by methods of this custom class.
    ui_base_class_name = ui_code_metadata['baseclass']

    # This base class or "None" if no such class exists.
    ui_base_class = getattr(QtWidgets, ui_base_class_name, None)

    # If no such base class exists, raise an exception.
    if ui_base_class is None:
        raise BetseeCacheException(
            title=QCoreApplication.translate(
                'convert_ui_to_py_file', 'PySide2 UI Compiler Error'),
            synopsis=QCoreApplication.translate(
                'convert_ui_to_py_file',
                'PySide2 widget base class "{0}" not found.'.format(
                    ui_base_class_name)))

    # Append an application-specific global to the Python code generated above,
    # declaring the sequence of all base classes that the main Qt window widget
    # class for this application *MUST* subclass (in order). In particular,
    # note that the base Qt widget class is subclassed last and thus remains
    # the "parent" base class of this multiple inheritance.
    ui_code_str_buffer.write('''
from PySide2.QtWidgets import {base_class}
{var_name} = ({form_class}, {base_class})
'''.format(
        var_name=BASE_CLASSES_GLOBAL_NAME,
        form_class=ui_form_class_name,
        base_class=ui_base_class_name,
    ))

    # Munge (i.e., modify) the Python code comprising the contents of the file
    # implementing this output module.
    ui_code_str = _munge_ui_code(
        # Original contents of this file.
        ui_code_str=ui_code_str_buffer.getvalue(),
        promote_obj_name_to_class=promote_obj_name_to_class,
    )

    # Write this Python code to this file, silently overwriting this file if
    # this file already exists.
    iofiles.write_str_to_filename(
        text=ui_code_str,
        output_filename=py_filename,
        is_overwritable=True,
    )


@type_check
def _munge_ui_code(
    ui_code_str: str,
    promote_obj_name_to_class: MappingType,
) -> str:
    '''
    Munge (i.e., modify) the passed string of Python code comprising the
    contents of the current ``.ui``-suffixed file to be subsequently written.

    Specifically, this function:

    * Prepends this code by a shebang line unambiguously running the executable
      binary for the active Python interpreter and machine architecture.
    * Globally replaces all lines of this code reducing vector SVG icons to
      non-vector in-memory pixmaps with lines preserving these icons as is. See
      this function's body for detailed commentary.
    * For the name of each instance variable of the main window and
      application-specific subclass to instantiate that variable to in the
      passed dictonary, replaces the single line of this code instantiating
      this variable to a non-promoted stock Qt widget class with a line instead
      instantiating this variable to this application-specific widget subclass.

    Parameters
    ----------
    ui_code_str : str
        String of Python code comprising the original contents of this file.
    promote_obj_name_to_class : MappingType
        See the :func:`convert_ui_to_py_file_if_able` function for comments.

    Returns
    ----------
    str
        String of Python code comprising the modified contents of this file.
    '''

    #FIXME: Consider filing an upstream PySide2 issue documenting this, ideally
    #with a minimal UI file as an example.
    #FIXME: Actually, we've done that now. When this issue is resolved, remove
    #this ad-hack kludge.

    # Globally replace all lines of this code reducing vector SVG icons to
    # non-vector in-memory pixmaps with lines preserving these icons as is.
    # This reduction has a variety of harmful side effects, including:
    #
    # * Preventing these icons from being upscaled. Qt unconditionally refuses
    #   to upscale in-memory pixmaps, which are of finite resolution and hence
    #   *NOT* upscalable without introducing spurious artifacts. Vector images
    #   are of infinite resolution and hence suffer no such issues. SVG icons
    #   should thus be upscalable, but (due to this reduction) are *NOT*. This
    #   issue is exacerbated by modern high-DPI displays and SVG icons
    #   rasterized at inappropriately small scale, rendering such icons
    #   effectively illegible.
    # * Pixellating these icons, even when rasterized at appropriate scale.
    #
    # For example, the QIcon created from a "lock_fill.svg" icon is as follows:
    #
    #     icon = QtGui.QIcon()
    #     icon.addPixmap(QtGui.QPixmap("://icon/open_iconic/lock_fill.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
    #
    # This QIcon should instead be created as:
    #
    #    # Note the need to pass an additional size, which remains
    #    # unconstrained due to the infinitely rescalable nature of vector
    #    # icons.
    #    icon = QtGui.QIcon()
    #    icon.addFile("://icon/open_iconic/lock_fill.svg", QtCore.QSize(), QtGui.QIcon.Normal, QtGui.QIcon.Off)
    #
    # Indeed, there no longer appears to be any benefit to *EVER* calling the
    # icon.addPixmap() versus icon.addFile() method. The latter suffices for all
    # filetypes and is thus always preferable -- regardless of filetype.
    ui_code_str = regexes.replace_substrs_line(
        text=ui_code_str,
        regex=(
            r'^(\s*icon\d*\.add)'
            r'Pixmap\(QtGui\.QPixmap\('
            r'("[^"]+\.svg")'
            r'\)(.*)$'
        ),
        replacement=r'\1File(\2, QtCore.QSize()\3',
    )

    # For the name of each instance variable of the main window and
    # application-specific subclass to instantiate that variable to...
    for promote_obj_name, promote_class in (
        promote_obj_name_to_class.items()):
        # If this name is *NOT* a valid Python identifier, raise an exception.
        # This is essential to avoid edge-case issues in which this name
        # erroneously contains regular expression syntax.
        pyident.die_unless_unqualified(promote_obj_name)

        # If this subclass is *NOT* a Qt object subclass, raise an exception.
        classes.die_unless_subclass(
            subclass=promote_class, superclass=QObject)

        # Unqualified name of this class.
        promote_class_name = classes.get_name(promote_class)

        # Fully-qualified name of the module defining this class.
        promote_class_module_name = classes.get_module_name(promote_class)

        # Append this code with a line importing this class.
        ui_code_str += 'from {} import {}\n'.format(
            promote_class_module_name, promote_class_name)

        # Regular expression matching the single line of this code
        # instantiating this variable to a non-promoted stock Qt widget class
        # into the following two numeric groups:
        #
        # 1. The assignment statement prefix for this instantiation.
        # 2. The parameters to be passed to this instantiation.
        promote_regex = (
            # Assignment statement.
            r'^(\s*self\.{promote_obj_name}\s*=\s*)'
            # Name of the non-promoted stock Qt widget class to be ignored.
            r'QtWidgets\.{widget_class_name}'
            # Parameters passed to this instantiation.
            r'(\([^)]*\))$'
        ).format(
            promote_obj_name=promote_obj_name,
            widget_class_name=IDENTIFIER_UNQUALIFIED_REGEX,
        )

        # Globally replace this line with a line instead instantiating this
        # variable to this application-specific widget subclass, implicitly
        # raising an exception if no such line exists.
        ui_code_str = regexes.die_unless_replace_substrs_line(
            text=ui_code_str,
            regex=promote_regex,
            replacement=r'\1{}\2'.format(promote_class_name),
        )

    # Prepend this code by a shebang running the active Python interpreter.
    ui_code_str = '{}\n{}'.format(pys.get_shebang(), ui_code_str)

    # Return this code.
    return ui_code_str



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Submodule providing general-purpose access to the :class:`QMainWindow`
singleton for this application.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication
from PySide2.QtWidgets import QMainWindow
from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.guiexception import BetseePySideWindowException

# ....................{ GLOBALS                            }....................
_MAIN_WINDOW = None
'''
Main window singleton widget for this application.
'''

# ....................{ GETTERS                            }....................
def get_main_window() -> QMainWindow:
    '''
    Singleton main window widget for this application if already instantiated
    by the :class:`betsee.gui.guimain.BetseeGUI` class *or* raise an exception
    otherwise (i.e., if this widget is uninstantiated).

    Design
    ----------
    To avoid circular import dependencies, this getter intentionally resides in
    this submodule known *not* to be subject to these dependencies rather than
    in an arguably more germane submodule known to be subject to these
    dependencies (e.g., :mod:`betsee.gui.window.guimainwindow`).

    Returns
    ----------
    QMainWindow
        This widget.

    Raises
    ----------
    BetseePySideWindowException
        If this widget has yet to be instantiated (i.e., if the
        :func:`set_main_window` function has yet to be called).
    '''

    # If this widget is uninstantiated. raise an exception.
    if _MAIN_WINDOW is None:
        raise BetseePySideWindowException(QCoreApplication.translate(
            'guiappwindow',
            'Main window singleton widget uninstantiated.'))

    # Else, this singleton has been instantiated. Return this singleton.
    return _MAIN_WINDOW

# ....................{ SETTERS                            }....................
@type_check
def set_main_window(main_window: QMainWindow) -> None:
    '''
    Set the main window singleton widget for this application.

    Parameters
    ----------
    main_window : QMainWindow
        Main window widget to set as this application's singleton.

    Raises
    ----------
    BetseePySideWindowException
        If this widget has already been instantiated (i.e., if the
        :class:`QApplication` singleton already defines the application-specific
        ``betsee_main_window`` attribute).
    '''

    # Globals modified below.
    global _MAIN_WINDOW

    # Log this attempt.
    logs.log_debug('Preserving main window...')

    # If this singleton is already instantiated. raise an exception.
    if _MAIN_WINDOW is not None:
        raise BetseePySideWindowException(QCoreApplication.translate(
            'guiappwindow',
            'Main window singleton widget already instantiated.'))

    # Set this global.
    _MAIN_WINDOW = main_window

# ....................{ UNSETTERS                          }....................
@type_check
def unset_main_window() -> None:
    '''
    Unset the main window singleton widget for this application.

    This function is intended to be called *only* on application destruction as
    a safety measure to avoid garbage collection issues.
    '''

    # Globals modified below.
    global _MAIN_WINDOW

    # Unset this global.
    _MAIN_WINDOW = None



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based stack widget page controllers specific to spatial settings.
'''

# ....................{ IMPORTS                            }....................
# from PySide2.QtCore import QCoreApplication #, Signal, Slot
from PySide2.QtWidgets import QMainWindow
from betse.science.parameters import Parameters
from betse.science.config.confenum import CellLatticeType
# from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC

# ....................{ SUBCLASSES                         }....................
class QBetseeSimConfSpaceStackedWidgetPager(QBetseeControllerABC):
    '''
    :mod:`PySide2`-based stack widget page controller, connecting all editable
    widgets of the spatial page with the corresponding low-level settings of the
    current simulation configuration.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def init(self, main_window: QMainWindow) -> None:

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Simulation configuration state object.
        sim_conf = main_window.sim_conf

        # Initialize all intracellular widgets on this page.
        main_window.sim_conf_space_intra_cell_radius.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.cell_radius)
        main_window.sim_conf_space_intra_lattice_disorder.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.cell_lattice_disorder)
        main_window.sim_conf_space_intra_lattice_type.init(
            sim_conf=sim_conf,
            sim_conf_alias=Parameters.cell_lattice_type,
            enum_member_to_widget_value={
                CellLatticeType.HEX: (
                    main_window.sim_conf_space_intra_lattice_hex),
                CellLatticeType.SQUARE: (
                    main_window.sim_conf_space_intra_lattice_square),
            },
        )

        # Initialize all extracellular widgets on this page.
        main_window.sim_conf_space_extra_grid_size.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.grid_size)
        main_window.sim_conf_space_extra_is_ecm.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.is_ecm)
        main_window.sim_conf_space_extra_world_len.init(
            sim_conf=sim_conf, sim_conf_alias=Parameters.world_len)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Application-specific exception hierarchy.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on application startup, the
# top-level of this module may import *ONLY* from submodules guaranteed to:
# * Exist, including standard Python and BETSEE modules. This does *NOT*
#   include BETSE modules, which are *NOT* guaranteed to exist at this point.
# * Never raise exceptions on importation (e.g., due to module-level logic).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from abc import ABCMeta

# ....................{ EXCEPTIONS ~ superclass           }....................
#FIXME: Introduce into the "betse.exceptions" submodule as a new
#"BetseVerboseException" base class from which all other exception classes
#defined by that submodule should eventually subclass.
class BetseeException(Exception, metaclass=ABCMeta):
    '''
    Abstract base class of all application-specific exceptions.

    Attributes
    ----------
    title : str
        Human-readable title associated with this exception, typically
        constrained to at most two to three words.
    synopsis : str
        Human-readable synopsis tersely describing this exception, typically
        constrained to a single sentence.
    exegesis : optional[str]
        Human-readable explanation fully detailing this exception if any,
        typically spanning multiple sentences, *or* ``None`` otherwise (i.e.,
        if no such explanation is defined).
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(
        self,
        synopsis: str,
        title: str = None,
        exegesis: str = None,
    ) -> None:
        '''
        Initialize this exception.

        Parameters
        ----------
        synopsis : str
            Human-readable synopsis tersely describing this exception,
            typically constrained to a single sentence.
        title : optional[str]
            Human-readable title associated with this exception, typically
            constrained to at most two to three words. Defaults to ``None``, in
            which case the title defaults to the translated string returned by
            the :meth:`_title_default` property.
        exegesis : optional[str]
            Human-readable explanation fully detailing this exception,
            typically spanning multiple sentences. Defaults to ``None``, in
            which case no such explanation is defined.
        '''

        # Since the @type_check decorator is unavailable at this early point in
        # application startup, manually assert the types of these parameters.
        assert isinstance(synopsis, str), '"{}" not a string.'.format(synopsis)
        assert isinstance(title, (str, type(None))), (
            '"{}" neither a string nor "None".'.format(title))
        assert isinstance(exegesis, (str, type(None))), (
            '"{}" neither a string nor "None".'.format(exegesis))

        # Initialize our superclass with the concatenation of this
        # exception's synopsis, a space, and exegesis (defaulting to the empty
        # string when unpassed).
        super().__init__(
            synopsis + (' ' + exegesis if exegesis is not None else ''))

        # If no title was explicitly passed...
        if title is None:
            # Attempt to fallback to the default title defined by this
            # exception subclass.
            try:
                title = self._title_default
            # If doing so fails due to PySide2 being unimportable and hence
            # uninstalled, fallback to an untranslated placeholder. Since
            # PySide2 is unavailable, this placeholder should never be used.
            except ImportError as exception:
                title = str(exception)

        # Classify all passed parameters.
        self.title = title
        self.synopsis = synopsis
        self.exegesis = exegesis

    # ..................{ PROPERTIES                        }..................
    @property
    def _title_default(self) -> str:
        '''
        Default human-readable title associated with *all* exceptions of this
        type for which no ``title`` parameter is passed at instantiation time.
        '''

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate('BetseeException', 'Horrible Error')

# ....................{ EXCEPTIONS ~ general              }....................
class BetseeCacheException(BetseeException):
    '''
    General-purpose exception applicable to user-specific caching, including
    dynamic generation of pure-Python modules imported at runtime.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseeCacheException', 'Cache Error')


class BetseeLibException(BetseeException):
    '''
    General-purpose exception applicable to all optional and mandatory
    third-party dependencies, including :mod:`PySide2`.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideException', 'Dependency Error')


class BetseeSimConfException(BetseeException):
    '''
    General-purpose exception applicable to simulation configuration state
    handling (e.g., whether a simulation configuration is currently open).
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseeSimConfException', 'Simulation Configuration Error')

# ....................{ EXCEPTIONS ~ psd                  }....................
class BetseePySideException(BetseeLibException):
    '''
    General-purpose exception applicable to :mod:`PySide2`, this application's
    principal third-party dependency.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideException', 'PySide2 Error')


class BetseePySideClipboardException(BetseePySideException):
    '''
    General-purpose exception applicable to all interaction with the
    platform-specific system clipboard.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideClipboardException', 'Clipboard Error'),


class BetseePySideFocusException(BetseePySideException):
    '''
    General-purpose exception applicable to all handling of interactive
    keyboard input focus for widgets.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideFocusException', 'Widget Focus Error')

# ....................{ EXCEPTIONS ~ psd : thread         }....................
class BetseePySideThreadException(BetseePySideException):
    '''
    :class:`PySide2.QtCore.QThread`-specific exception.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideThreadException', 'Thread Error')

# ....................{ EXCEPTIONS ~ psd : thread: worker }....................
class BetseePySideThreadWorkerException(BetseePySideThreadException):
    '''
    Multithreaded worker object-specific exception, where "worker" implies any
    :class:`QObject`- or :class:`QRunnable`-derived object isolated in whole or
    part to a secondary application thread.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideThreadWorkerException', 'Thread Worker Error')


class BetseePySideThreadWorkerStopException(BetseePySideThreadWorkerException):
    '''
    Multithreaded worker object-specific exception internally raised by the
    ``_halt_work_if_requested`` methods and caught by the ``start`` methods
    defined on these objects.

    This exception is intended exclusively for private use by the
    aforementioned methods as a crude (albeit sufficient) means of facilitating
    superclass-subclass intercommunication.
    '''

    pass

# ....................{ EXCEPTIONS ~ psd : widget         }....................
class BetseePySideWidgetException(BetseePySideException):
    '''
    General-purpose exception applicable to :mod:`PySide2` widgets.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideWidgetException', 'Widget Error')


class BetseePySideApplicationException(BetseePySideWidgetException):
    '''
    :class:`PySide2.QtWidgets.QApplication`-specific exception pertaining to
    the application singleton for this... application.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideApplicationException', 'Singleton Error')


class BetseePySideMenuException(BetseePySideWidgetException):
    '''
    :class:`PySide2.QtWidgets.QMenu`-specific exception.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideMenuException', 'Menu Error')


class BetseePySideMessageBoxException(BetseePySideWidgetException):
    '''
    :class:`PySide2.QtWidgets.QMessageBox`-specific exception.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideMessageBoxException', 'Message Box Error')


class BetseePySideSpinBoxException(BetseePySideWidgetException):
    '''
    General-purpose exception applicable to all concrete
    :class:`PySide2.QtWidgets.QAbstractSpinBox` widgets (e.g.,
    :class:`PySide2.QtWidgets.QDoubleSpinBox` widgets).
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideSpinBoxException', 'Spin Box Error')


class BetseePySideTreeWidgetException(BetseePySideWidgetException):
    '''
    :class:`PySide2.QtWidgets.QTreeWidget`-specific exception.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideTreeWidgetException', 'Tree Widget Error')


class BetseePySideWindowException(BetseePySideWidgetException):
    '''
    :class:`PySide2.QtWidgets.QMainWindow`-specific exception.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideWindowException', 'Window Error')

# ....................{ EXCEPTIONS ~ psd : widget : enum  }....................
#FIXME: Rename to "BetseePySideEnumWidgetException" for orthogonality.
class BetseePySideWidgetEnumException(BetseePySideWidgetException):
    '''
    General-purpose exception applicable to mutually exclusive :mod:`PySide2`
    widgets typically converted to and from lower-level enumeration members.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideWidgetEnumException', 'Enumerable Widget Error')


class BetseePySideComboBoxException(BetseePySideWidgetEnumException):
    '''
    :class:`PySide2.QtWidgets.QComboBox`-specific exception.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideComboBoxException', 'Combo Box Error')


class BetseePySideRadioButtonException(BetseePySideWidgetEnumException):
    '''
    :class:`PySide2.QtWidgets.QRadioButton`-specific exception.
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideRadioButtonException', 'Radio Button Error')

# ....................{ EXCEPTIONS ~ psd : widget: betsee }....................
class BetseePySideEditWidgetException(BetseePySideException):
    '''
    General-purpose exception applicable to application-specific editable
    widgets (i.e., instances of the
    :mod:`betsee.util.widget.abc.guiwdgabc.QBetseeObjectMixin` superclass).
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseePySideEditWidgetException', 'Editable Widget Error')

# ....................{ EXCEPTIONS ~ simmer               }....................
class BetseeSimmerException(BetseePySideException):
    '''
    General-purpose exception applicable to the **simulator** (i.e.,
    :mod:`PySide2`-based object both displaying *and* controlling the execution
    of simulation phases).
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseeSimmerException', 'Simulator Error')


class BetseeSimmerBetseException(BetseeSimmerException):
    '''
    General-purpose exception intended to encapsulate *all* low-level
    exceptions raised by BETSE simulations (e.g., computational instabilities).
    '''

    @property
    def _title_default(self) -> str:

        # Defer heavyweight imports *NOT* guaranteed to exist.
        from PySide2.QtCore import QCoreApplication

        # Translate this title.
        return QCoreApplication.translate(
            'BetseeSimmerBetseException', 'BETSE Error')
